/**
   @file    EnergyConsumption.cc
   @ingroup EnergyProperties
   @brief   Energy consumption property
   @author  Robert Mijakovic
   @verbatim
    Revision:       $Revision$
    Revision date:  $Date$
    Committed by:   $Author$

    This file is part of the Periscope Tuning Framework.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2014, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include "EnergyConsumption.h"
#include <iostream>

#include "global.h"
#include "Metric.h"
#include "Context.h"
#include "PropertyID.h"
#include "psc_errmsg.h"
#include <sstream>

//Description :  Tells you the energy consumption of code region
//ENOPT Formula:  PSC_ENOPT_ALL_CORES
//Recommendation :
//


const std::string EnergyConsumption::name_constant = "Energy Consumption";


EnergyConsumption::EnergyConsumption( Context* ct, Context* phaseCt ) : Property( ct ) {
    phaseCycles       = 0;
    context           = ct;
    phaseContext      = phaseCt;
    energyConsumption = 0;
}


PropertyID EnergyConsumption::id() {
    return ENERGY_CONSUMPTION;
}


void EnergyConsumption::print() {
    std::cout << "Property:" << name()
              << "  Process " << context->getRank()
              << "  Thread " << context->getThread() << std::endl
              << "                  " << context->getRegion()->str_print() << std::endl;
}


bool EnergyConsumption::condition() const {
    return energyConsumption >= 0;
}


double EnergyConsumption::confidence() const {
    return 1.0;
}


double EnergyConsumption::severity() const {
    return energyConsumption;
}


Context* EnergyConsumption::get_phaseContext() {
    return phaseContext;
}


Gather_Required_Info_Type EnergyConsumption::request_metrics() {
    Context* ct = new Context( context->getRegion(), context->getRank(), 0 );
    pdb->request( ct, PSC_ENOPT_ALL_CORES );

    return ALL_INFO_GATHERED;
}


std::string EnergyConsumption::name() {
    return name_constant;
}


void EnergyConsumption::evaluate() {
    Context* ct  = new Context( context->getRegion(), context->getRank(), 0 );
    INT64    raw = pdb->get( ct, PSC_ENOPT_ALL_CORES );
    energyConsumption = static_cast<double>( raw ) / 1e9;      // NOTE: Score-P returns nano-Joules
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ), "Got energy consumption: %f\n", energyConsumption );

    delete ct;
}


std::string EnergyConsumption::info() {
    std::stringstream stream;
    stream << '\t' << " ENERGY_CONSUMPTION: " << energyConsumption;
    return stream.str();
}


Property* EnergyConsumption::clone() {
    return new EnergyConsumption( context, phaseContext );
}
