/**
   @file    Tuning.cc
   @ingroup TuningStrategy
   @brief   Autotune search strategy
   @author  Michael Gerndt
   @verbatim
    Revision:       $Revision$
    Revision date:  $Date$
    Committed by:   $Author$

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2014, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include "global.h"
#include "Tuning.h"
#include "ExecTime.h"
#include "EnergyConsumption.h"
#include <iostream>
#include "psc_errmsg.h"
#include <analysisagent.h>
#include "selective_debug.h"
#include "TuningParameter.h"
#ifdef __p575
#include <strings.h>
#endif

#include "ScorepMPIStrategy.h"
#include "OpenMPAnalysisScoreP.h"
#include "ConfigAnalysis.h"

#include <boost/foreach.hpp>

#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>

/**
 * @brief Constructor
 * @ingroup TuningStrategy
 *
 * @param reg       Tuned region
 * @param scenarios A list of scenarios that has to evaluated
 * @param request   Strategy request containing the name of per-experiment analysis strategy
 * @param pedantic  Pedantic feature is switched on, i.e. are all properties going to be reported.
 **/
TuningStrategy::TuningStrategy( Region* reg, const std::list<Scenario*>* scenarios,
                                StrategyRequest* request, bool pedantic ) :
    Strategy( pedantic ) {
    tunedRegion      = reg;
    scenarioList     = scenarios;
    analysisStrategy = NULL;

    if( agent->get_leader() && psc_get_debug_level() >= 3 ) {
        std::list<Scenario*>::const_iterator scenario;
        for( scenario = scenarioList->begin(); scenario != scenarioList->end(); scenario++ ) {
            printf( "TuningStrategy scenario:\n" );
            ( *scenario )->print();
            fflush( stdout );
        }
    }

    map_scenarios_to_ranks();

    if( !analysisStrategy ) {
        if( request && strcmp( request->getGeneralInfo()->strategy_name.c_str(), "MPI" ) == 0 ) {
            analysisStrategy = new ScorepMPIStrategy( appl, request->getGeneralInfo()->pedantic );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ),
                        "TuningStrategy has loaded the MPI sub-strategy.\n" );
        }
        else if( request && strcmp( request->getGeneralInfo()->strategy_name.c_str(), "OMP" ) == 0 ) {
            analysisStrategy = new OpenMPAnalysisScoreP( request->getGeneralInfo()->pedantic );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ),
                        "TuningStrategy has loaded the OMP sub-strategy.\n" );
        }
        else if( request && strcmp( request->getGeneralInfo()->strategy_name.c_str(), "ConfigAnalysis" ) == 0 ) {
            analysisStrategy = new ConfigAnalysis( request->getConfiguration().configuration_union.PropertyRequest_list,  request->getGeneralInfo()->pedantic );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ),
                        "TuningStrategy has loaded the configurable analysis sub-strategy.\n" );
        }
        else {
            analysisStrategy = NULL;
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ),
                        "TuningStrategy has no analysis sub-strategy specified.\n" );
        }
    }
}

/**
 * @brief Requests tuning actions from the MRI monitor
 * @ingroup TuningStrategy
 *
 * Requests dynamic tuning actions from the MRI monitor.
 *
 */
void TuningStrategy::requestTuningActions() {
    std::map<int, std::list<Scenario*> >::iterator sc_list_it;
    std::list<Scenario*>::iterator                 sc_it;
    std::list<TuningSpecification*>*               ts_list;
    std::list<TuningSpecification*>::iterator      ts_it;
    for( sc_list_it = TS_related_scenarios_per_rank.begin(); sc_list_it != TS_related_scenarios_per_rank.end(); sc_list_it++ ) {
        int rank = sc_list_it->first;
        for( sc_it = sc_list_it->second.begin(); sc_it != sc_list_it->second.end(); sc_it++ ) {
            ts_list = ( *sc_it )->getTuningSpecifications();
            if( !ts_list ) {
                continue;
            }
            for( ts_it = ts_list->begin(); ts_it != ts_list->end(); ts_it++ ) {
                if( processInRanks( ( *ts_it )->getRanks(), rank ) ) {
                    dp->addTuningRequest( rank, ( *ts_it ) );
                }
            }
        }
    }
}

/**
 * @brief Creates a candidate property list for the first run.
 *
 * @ingroup TuningStrategy
 *
 * Creates a candidate property list for the first run and prepares requests
 * for runtime information. The requested property list is consisted of requests
 * of a particular per-experiment analysis strategy and tuning objectives.
 *
 * @param r Region for which properties are requested
 * @return Returns true as this is single-step strategy
 */
bool TuningStrategy::reqAndConfigureFirstExperiment( Region* r ) {
// TRUE can start; FALSE not ready

    phaseRegion = r;

    std::list<ApplProcess>                      controlled_processes = dp->get_controlled_processes();
    std::list<ApplProcess>::iterator            process;
    std::list<Scenario*>::const_iterator        scenario;
    int                                         scenarioId;
    const std::list<PropertyRequest*>*          propertyReqList;
    std::list<PropertyRequest*>::const_iterator propertyReq;
    Prop_List::iterator                         prop_it;

    std::map<int, std::list<Scenario*> >::iterator PRs_it;

    // Requests objective properties for a tuned region - For this properties purpose is set to PURPOSE_TUNING
    for( PRs_it = PR_related_sceanrios_per_rank.begin(); PRs_it != PR_related_sceanrios_per_rank.end(); PRs_it++ ) {
        int rank = PRs_it->first;
        for( scenario = PRs_it->second.begin(); scenario != PRs_it->second.end(); scenario++ ) {
            scenarioId      = ( *scenario )->getID();
            propertyReqList = ( *scenario )->getPropertyRequests();
            for( propertyReq = propertyReqList->begin(); propertyReq != propertyReqList->end(); propertyReq++ ) {
                if( processInRanks( ( *propertyReq )->getRanks(), rank ) ) {
                    createCandidatePropertyList( ( *scenario )->getRegion(), scenarioId, rank, *propertyReq );
                }
            }
        }
    }

    // Request tuning actions
    requestTuningActions();

    if( psc_get_debug_level() >= 3 ) {
        agent->print_property_set( candProperties, "SET OF CANDIDATE PROPERTIES", false, false );
    }
    if( agent->get_leader() ) {
        psc_dbgmsg( 3, "Tuning Strategy Step %d\n", strategy_steps );
    }

    for( prop_it = candProperties.begin(); prop_it != candProperties.end(); prop_it++ ) {
        ( *prop_it )->request_metrics();
    }

    // Runs other analysis strategies's configure first experiment, i.e.,
    // per-experiment analysis, and prepares requests for runtime information
    if( analysisStrategy ) {
        analysisStrategy->reqAndConfigureFirstExperiment( r );
    }

    return true;
}

/**
 * @brief Configures next experiment.
 *
 * @ingroup TuningStrategy
 *
 * Requests tuning actions from tuning strategies and sends them to the MRI monitor.
 *
 */
void TuningStrategy::configureNextExperiment() {
    psc_dbgmsg( 3, "Tuning Strategy configureNextExperiment\n" );

    dp->transfer_requests_to_processes();

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ), "configureNextExperiment done\n" );
}

/**
 * @brief Evaluates the properties from the previous and requests new experiment.
 *
 * @ingroup TuningStrategy
 *
 * Evaluates the properties from the candidate list and moves them to the found
 * list. Found properties are then marked with their purpose for the frontend.
 *
 * @return Decision on continuation of tuning analysis strategy.
 */
bool TuningStrategy::evaluateAndReqNextExperiment() {
    std::list<Scenario*>::const_iterator scenario;
    int                                  scenarioId;


    Prop_List::iterator prop_it;
    double              max_severity = 0;
    bool                cont         = 0;

    //if (psc_get_debug_level() >= 2)
    //agent->print_property_set(candProperties, strdup("SET OF CANDIDATE PROPERTIES"), false, true);

    //Evaluate candidate properties
    psc_dbgmsg( 3, "%s: analyzing results... %d candidate properties.\n", name().c_str(), candProperties.size() );

    for( prop_it = candProperties.begin(); prop_it != candProperties.end(); prop_it++ ) {
        ( *prop_it )->evaluate();
        if( ( *prop_it )->condition() ) {
            /*
             * Below comment is wrong! All candidate properties created by this strategy
             * have purpose tuning. If not then print an error message
             */
            // Found properties that do not belong to tuning are marked as analysis properties
            if( ( *prop_it )->get_Purpose() != PSC_PROPERTY_PURPOSE_TUNING ) {
                psc_errmsg( "A candidate property %s of the Tuning strategy has purpose which is not tuning!" );
            }

            foundProperties.push_back( *prop_it );
        }
    }

    if( analysisStrategy ) {
        cont = analysisStrategy->evaluateAndReqNextExperiment();
    }

    /*
     * foundProperties was changed to be a global list which is pushed by all strategies
     * Here we go over the list of all found properties (the ones created by this tuning strategy,
     * its analysis strategy as well as potentially *WARNING* other strategies running on this
     * agent) and assign a ScenarioID as well as its purpose to be analysis)
     */
    for( prop_it = foundProperties.begin(); prop_it != foundProperties.end(); prop_it++ ) {
        if( ( *prop_it )->get_Purpose() != PSC_PROPERTY_PURPOSE_TUNING ) {
            ( *prop_it )->set_Purpose( PSC_PROPERTY_PURPOSE_ANALYSIS );
            std::list<Scenario*>::const_iterator scenario;
            std::list<Scenario*>                 scenario_list = PR_related_sceanrios_per_rank[ ( *prop_it )->get_rank() ];
            for( scenario = scenario_list.begin(); scenario != scenario_list.end(); scenario++ ) {
                ( *prop_it )->add_ScenarioId( ( *scenario )->getID() );
            }
        }
    }

    if( psc_get_debug_level() >= 2 ) {
        agent->print_property_set( foundProperties, strdup( "SET OF FOUND PROPERTIES" ), true, true );
    }

    return cont;
}

/**
 * @brief Returns the name of the tuning strategy.
 *
 * @ingroup TuningStrategy
 *
 */
std::string TuningStrategy::name() {
    return "TuningStrategy";
}

/**
 * @brief Creates a candidate property list.
 *
 * Creates a candidate property list from a tuned region and property requests.
 * Marks properties with scenario id and purpose.
 */
void TuningStrategy::createCandidatePropertyList( Region* region, int scenarioId, int processRank, PropertyRequest* propertyRequest ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAgentStrategy ), "Creating an objective for scenario %d (rank:%d)\n", scenarioId, processRank );
    Context* context = new Context( phaseRegion, processRank, 0 );

    BOOST_FOREACH( int propertyId, *propertyRequest->getPropertyIDs() ) {
        Property* prop;

        switch( propertyId ) {
        case EXECTIME:
            prop = new ExecTimeProp( new Context( region, processRank, 0 ), context, 0.0 );
            break;

        case ENERGY_CONSUMPTION:
            prop = new EnergyConsumption( new Context( region, processRank, 0 ), context );
            break;

        default:
            prop = NULL;
            psc_errmsg( "Unknown objective in the tuning strategy: %i\n", propertyId );
            abort();
        }

        prop->add_ScenarioId( scenarioId );
        prop->set_Purpose( PSC_PROPERTY_PURPOSE_TUNING );
        candProperties.push_back( prop );
    }
}

void TuningStrategy::map_scenarios_to_ranks() {
    std::list<Scenario*>::const_iterator scenario;
    std::list<ApplProcess>::iterator     process;
    std::list<ApplProcess>               appl_processes = dp->get_controlled_processes();
    for( process = appl_processes.begin(); process != appl_processes.end(); process++ ) {
        int                  rank = process->rank;
        std::list<Scenario*> empty_list;
        TS_related_scenarios_per_rank[ rank ] = empty_list;
        PR_related_sceanrios_per_rank[ rank ] = empty_list;

        if( psc_get_debug_level() >= 8 ) {
            std::cout << "Application process rank " << rank << " has following scenarios:" << endl;
        }

        for( scenario = ( *scenarioList ).begin(); scenario != ( *scenarioList ).end(); scenario++ ) {
            if( psc_get_debug_level() >= 8 ) {
                std::cout << "  Scenario " << ( *scenario )->getID() << " as ";
            }
            if( ( *scenario )->getRankAffectedBySceanrioTS( rank ) ) {
                if( psc_get_debug_level() >= 8 ) {
                    std::cout << "  a tuning specification ";
                }
                TS_related_scenarios_per_rank[ rank ].push_back( *scenario );
            }
            if( ( *scenario )->getRankAffectedBySceanrioPR( rank ) ) {
                if( psc_get_debug_level() >= 8 ) {
                    std::cout << "  a property request";
                }
                PR_related_sceanrios_per_rank[ rank ].push_back( *scenario );
            }
            if( psc_get_debug_level() >= 8 ) {
                std::cout << endl;
            }
        }
    }

    /*
     * The below block of code is not needed
     */
    //If the configurable strategy is requested, properties are created but the properties are not marked
    //with a scenario. Here, we use as default the first scenario. May be we should use all of them.
//  for (process = controlled_processes.begin(); process != controlled_processes.end(); process++) {
//	  if (process->getScenariosPerTuningSpecification()->size()==0){
//		  Scenario *sc= *scenarios->begin();
//		  process->addScenarioPerTuningSpecification(sc);
//	  }
//	  if (process->getScenariosPerPropertyRequest()->size()==0){
//		  Scenario *sc= *scenarios->begin();
//		  process->addScenarioPerPropertyRequest(sc);
//	  }
//  }
}
