/**
   @file    MPIParametersPlugin.h
   @ingroup MPIParametersPlugin
   @brief   Demo Plugin
   @author  Toni Pimenta, Eduardo César, Anna Sikora
   @verbatim
    Revision:       $Revision$
    Revision date:  $Date$
    Committed by:   $Author$

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2014, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

/**
   @defgroup MPIParametersPlugin MPI Parameters Plugin
   @ingroup AutotunePlugins
 */

#ifndef MPIPARAMETERS_PLUGIN_H_
#define MPIPARAMETERS_PLUGIN_H_

#include "AutotunePlugin.h"
#include "ISearchAlgorithm.h"

enum intervalOperator {
    SUM, MUL, POW
};

enum mpiImplementationName {
    GENERIC, OPENMPI, IBM, INTEL, unknown, SEARCHALG, AUTOMATICEAGER, INDKEEP, GDE3POPSZ, SAMPCOUNT
};

/**
 * Stores the parameter information extracted from the configuration file
 * interval: start-step-end
 * name of the MPI parameter
 * values: vector of possible values as strings
 */

struct MPIParameterTP {
    mpiImplementationName MPIType;
    std::string           env_var_name;
    long                  start;
    int                   step;
    long                  end;
    bool                  ranged;
    std::vector<string>   values;
    intervalOperator      op;
    MPIParameterTP() : start( 0 ), step( 0 ), end( 0 ), ranged( false ), op( SUM ) {
    }
};




class MPIParametersPlugin : public IPlugin {
    vector<TuningParameter*>    tuningParameters;                   ///< Vector of tuning parameters
    mpiImplementationName       MPIType;                            ///< To differenciate MPI implementations (IBM, OpenMPI, Intel, others)
    vector<MPIParameterTP>      conftps;                            ///< Vector of configuration parameters parsed from the configuration file
    string                      exports;                            ///< For those cases (others) where parameters are set through environments variables
    string                      restartCommand;                     ///< For those cases (IBM, INTEL, OpenMPI) where parameters are set in the application command line
    string                      appflags;
    string                      automatic_eager, automatic_buf_mem; ///< For storing the names of the eager limit and buf mem parameters when their ranges are automatically determined by the plugin
    bool                        instrumented_flag;
    map< long, vector<string> > valuesMap;                          ///< Maps integers to string values for some parameter

    VariantSpace variantSpace;
    SearchSpace  searchSpace;

    ISearchAlgorithm* searchAlgorithm;
    Region*           tunedRegion;

    void extractTuningParametersFromConfigurationFile();

public:
    void initialize( DriverContext*   context,
                     ScenarioPoolSet* pool_set );

    bool analysisRequired( StrategyRequest** strategy );

    void startTuningStep( void );

    void createScenarios( void );

    void prepareScenarios( void );

    void defineExperiment( int               numprocs,
                           bool&             analysisRequired,
                           StrategyRequest** strategy );

    bool restartRequired( std::string& env,
                          int&         numprocs,
                          std::string& command,
                          bool&        is_instrumented );

    bool searchFinished( void );

    void finishTuningStep( void );

    bool tuningFinished( void );

    Advice* getAdvice( void );

    void finalize( void );

    void terminate( void );

    const int getNumParameters() const {
        return tuningParameters.size();
    }

    const int getNumValues( int i ) const {
        return ( tuningParameters.at( i )->getRangeTo() - tuningParameters.at( i )->getRangeFrom()
                 + tuningParameters.at( i )->getRangeStep() ) / tuningParameters.at( i )->getRangeStep();
    }

    const string getExportsString() const {
        return exports;
    }
};

vector<MPIParameterTP>getTuningParameters( const char*            filename,
                                           mpiImplementationName& MPItype );

#endif
