/**
   @file    MPIParametersPlugin.cc
   @ingroup MPIParametersPlugin
   @brief   MPI Parameters Plugin
   @author  Toni Pimenta, Gertvjola Saveta, Eduardo César, Anna Sikora
   @verbatim
    Revision:       $Revision$
    Revision date:  $Date$
    Committed by:   $Author$

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2014, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include "IndividualSearch.h"
#include "RandomSearch.h"
#include "GDE3Search.h"
#include "LoadSearch.h"

#include "MPIParametersPlugin.h"
#include <sstream>

/**
 * @brief Extracting Tuning Parameters from the Configuration File
 * @ingroup MPIParametersPlugin
 */
void MPIParametersPlugin::extractTuningParametersFromConfigurationFile() {
    // Should get filename from env var.
    char const* param_spec_file = getenv( "PSC_PARAM_SPEC_FILE" );
    if( param_spec_file != NULL ) {
        conftps = getTuningParameters( param_spec_file, MPIType );
    }
    else {
        conftps = getTuningParameters( "./param_spec.conf", MPIType );
    }

    if( conftps.empty() ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ),
                    "[#### AUTOTUNE ####]: No tuning parameters found in specification file file, set PSC_PARAM_SPEC_FILE to correct file path. Exiting.\n" );
        throw PTF_PLUGIN_ERROR( TUNING_PARAMETERS_NOT_FOUND );
    }
}

/**
 * @brief Initializes the MPIParameters tuning plugin.
 * @ingroup MPIParametersPlugin
 *
 * The tuning space is created. Here the configuration file is loaded and the
 * corresponding tuning parameters  are generated from the specified MPI
 * parameters. Then, the search strategy is chosen and the corresponding search
 * algorithm is instantiated.
 * In case the user indicates in the configuration file that the eager limit and buffer mem
 * parameters range of values are to be automatically determined by the plugin, the names of the
 * parameters are stored in automatic_eager and automatic_buf_mem members. Their corresponding tuning
 * parameters will be determined later after calling a Periscope pre-analysis.
 *
 * @param context a pointer to a context for a plugin
 * @param pool_set a pointer to a set of pools for a plugin
 *
 */
void MPIParametersPlugin::initialize( DriverContext*   context,
                                      ScenarioPoolSet* pool_set ) {
    MPIParameterTP mpisearch;
    int            ind_keep   = 0;
    int            gde_pop_sz = 0;
    int            scount     = 2;

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to initialize()\n" );
    this->context  = context;
    this->pool_set = pool_set;

    extractTuningParametersFromConfigurationFile();

    Application& app = Application::instance();

    list<Region*> regions = app.get_regions();

    if( regions.empty() ) {
        psc_abort( "MPIParametersPlugin: No code regions are known! Exiting.\n" );
    }

    Region* pr = app.get_phase_region();

    if( pr == NULL ) {
        pr = new Region();
        RegionIdent ri;

        ri.file_id = -1;
        ri.rfl     = -1;
        pr->set_ident( ri );
    }
    else {
        cout << "Printing region (phase):" << endl;
        pr->print();
    }
    tunedRegion = pr;

    for( int i = 0; i < conftps.size(); i++ ) {
        MPIParameterTP mpitp;

        mpitp = conftps.at( i );
        if( mpitp.MPIType == SEARCHALG ) {
            mpisearch = mpitp;
            continue;
        }
        if( mpitp.MPIType == AUTOMATICEAGER ) {
            automatic_eager   = mpitp.values.at( 0 );
            automatic_buf_mem = mpitp.values.at( 1 );
            continue;
        }
        if( mpitp.MPIType == INDKEEP ) {
            ind_keep = atoi( mpitp.values.at( 0 ).c_str() );
            continue;
        }
        if( mpitp.MPIType == GDE3POPSZ ) {
            gde_pop_sz = atoi( mpitp.values.at( 0 ).c_str() );
            continue;
        }
        if( mpitp.MPIType == SAMPCOUNT ) {
            scount = atoi( mpitp.values.at( 0 ).c_str() );
            continue;
        }

        TuningParameter* tp;

        tp = new TuningParameter();
        tp->setId( i );
        tp->setName( mpitp.env_var_name.c_str() );
        tp->setPluginType( MPI );
        tp->setRuntimeActionType( TUNING_ACTION_NONE );
        tp->setRange( mpitp.start, mpitp.end, mpitp.step );
        Restriction* res = new Restriction();
        res->setRegion( pr );
        res->setRegionDefined( true );
        tp->setRestriction( res );

        tuningParameters.push_back( tp );
        if( mpitp.ranged ) {
            std::vector<string> values;
            valuesMap.insert( make_pair( tp->getId(), values ) );
        }
        else {
            valuesMap.insert( make_pair( tp->getId(), mpitp.values ) );
        }
    }

    if( active_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ) ) > 0 ) {
        printf( "\nAll created tuning parameters\n=======================\n" );
        for( int i = 0; i < tuningParameters.size(); i++ ) {
            string str = tuningParameters[ i ]->toString();
            cout << str <<  "=====================" << endl;
        }
    }


    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "obtain getSearchInstance\n" );
    int    major, minor;
    string name, description;
    string selected_search_string;
    if( !mpisearch.values.empty() ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "Selecting search algorithm: %s\n", mpisearch.values.at( 0 ).c_str() );
        selected_search_string = string( mpisearch.values.at( 0 ) );
        context->loadSearchAlgorithm( selected_search_string, &major, &minor, &name, &description );
        searchAlgorithm = context->getSearchAlgorithmInstance( selected_search_string );
        if( selected_search_string.compare( "individual" ) == 0 ) {
            IndividualSearch* alg = reinterpret_cast<IndividualSearch*>( searchAlgorithm );
            if( ind_keep ) {
                alg->setScenariosToKeep( ind_keep );
            }
        }
        if( selected_search_string.compare( "random" ) == 0 ) {
            RandomSearch* alg = reinterpret_cast<RandomSearch*>( searchAlgorithm );
            alg->setSampleCount( scount );
        }
        if( selected_search_string.compare( "gde3" ) == 0 ) {
            GDE3Search* alg = reinterpret_cast<GDE3Search*>( searchAlgorithm );
            if( gde_pop_sz ) {
                alg->setPopulationSize( gde_pop_sz );
            }
        }
        if( selected_search_string.compare( "load" ) == 0 ) {
            LoadSearch* alg = reinterpret_cast<LoadSearch*>( searchAlgorithm );
        }
    }
    else {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "Selecting default search algorithm: exhaustive\n" );
        context->loadSearchAlgorithm( "exhaustive", &major, &minor, &name, &description );
        searchAlgorithm = context->getSearchAlgorithmInstance( "exhaustive" );
    }

    if( searchAlgorithm != NULL ) {
        print_loaded_search( major, minor, name, description );
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "Before Search algorithm init\n" );
        searchAlgorithm->initialize( context, pool_set );
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "After Search algorithm init\n" );
    }
    else {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "NULL pointer in searchAlgorithm\n" );
        throw PTF_PLUGIN_ERROR( NULL_REFERENCE );
    }

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "searchAlgorithm instance obtained\n" );
}

/**
 * @brief Defines a pre-analysis strategy.
 *
 * If the user has indicated that the range of values for the eager limit and buffer mem parameters
 * must be decided automatically by the plugin then a Configurable Analysis strategy is used for
 * requiring the EAGERLIMITDEPENDENT property-
 *
 * @ingroup MPIParametersPlugin
 * @return False if the user does not indicates to use the automatic eager limit range determination
 * and true otherwise.
 */
bool MPIParametersPlugin::analysisRequired( StrategyRequest** strategy ) {
    if( context->applUninstrumented() ) {
        return false;
    }

    if( !automatic_eager.empty() ) {
        list<PropertyRequest*>*     reqList                     = new list<PropertyRequest*>;
        StrategyRequestGeneralInfo* analysisStrategyGeneralInfo = new StrategyRequestGeneralInfo;

        analysisStrategyGeneralInfo->strategy_name     = "ConfigAnalysis";
        analysisStrategyGeneralInfo->pedantic          = true;
        analysisStrategyGeneralInfo->delay_phases      = 0;
        analysisStrategyGeneralInfo->delay_seconds     = 0;
        analysisStrategyGeneralInfo->analysis_duration = 1;

        PropertyRequest* req = new PropertyRequest();
        req->addPropertyID( EAGERLIMITDEPENDENT );
        req->addRegion( appl->get_main_region() );
        req->addAllProcesses();

        reqList->push_back( req );

        *strategy = new StrategyRequest( reqList, analysisStrategyGeneralInfo );
        return true;
    }
    return false;
}

/**
 * @brief Operations to be done at the start of a tuning step.
 *
 * @ingroup MPIParametersPlugin
 *
 */
void MPIParametersPlugin::startTuningStep( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to startTuningStep() \n" );
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: Create a VariantSpace from the tuning parameters.\n" );
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: Search prepared.\n" );
}

/**
 * @brief The Created Scenario Pool (csp) is populated here.
 *
 * If the user has not indicated to automatically determine the range of values for the eager limit
 * and buffer mem parameters this method simply creates the variant space using the tuning parameters
 * created at @see initialize and makes the search algorithm to create the scenarios.
 * If, on the contrary, the user has indicated to automatically determine the range of values for
 * the eager limit and buffer mem parameters then the measurements done in the Periscope pre-analysis
 * for the EAGERLIMITDEPENDANT property are retrieved. These measurement include the total number of bytes
 * communicated (mpi_msg_p2p_tot) and the total number of bytes communicated with messages of size
 * smaller than 64Kbytes (mpi_msg_p2p_thr), i.e., within the eager limit possible range.
 * If mpi_msg_p2p_thr represents less than 30% of the total communication volume the impact of this
 * parameter on the application performance is considered non relevant and it is not taken into
 * consideration. Consequently, only the parameters created at @see initialize will be taken into
 * consideration.
 * If, on the contrary, mpi_msg_p2p_thr represents 30% or more of the total communication then the
 * plugin will use the rest of the measurements obtained in the pre-analysis for determining the range
 * for the eager limit and buffer mem parameters. These measurements are the number of messages in
 * the ranges 1 to 2 Kbytes (mpi_msg_freq_2k), 2 to 4 Kbytes (mpi_msg_freq_4k), 4 to 8 Kbytes
 * (mpi_msg_freq_8k), 8 to 16 Kbytes (mpi_msg_freq_16k), 16 to 32 Kbytes (mpi_msg_freq_32k), and
 * 32 to 64 Kbytes (mpi_msg_freq_64k). The plugin determines the proportion of messages in each of
 * these intervals and proceeds to calculate the range in which most of the messages are into.
 * Initially the eager limit scenario interval is set from 1 to 64 Kbytes (the whole validity range).
 * If one of these intervals contains 50% or more of the total number of messages the initial limits
 * for the eager limit scenario range are set to the interval limits.
 * Then, the algorithm checks if two consecutive intervals contain more than 60% of the messages and,
 * if true, it sets the scenario limits to the lower limit of the first interval and the upper limit
 * of the second.
 * Next, it checks if three consecutive intervals contain more than 80% of the messages or if four
 * contain 90% of the messages, setting the scenario limits accordingly.
 * The step in the determined range is set depending on the size of the range determined.
 * Finally, once the eager limit range has been determined, the plugin uses the expression
 * limit_buf_mem = 2*limit_eager_limit for determining the limits of the buffer mem parameter and also
 * its step.
 * Once the ranges of both parameters have been determined, the corresponding tuning parameters are
 * created and added to the set of tuning parameters defined at @see initialize and the scenarios
 * created in the search algorithm.
 *
 * @ingroup MPIParametersPlugin
 */
void MPIParametersPlugin::createScenarios( void ) {
    static bool pass = false;

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to createScenarios()\n" );
    if( searchAlgorithm == NULL ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "Search algorithm not instantiated\n" );
        throw PTF_PLUGIN_ERROR( NULL_REFERENCE );
    }

    if( !pass ) {
        //psc_dbgmsg(PSC_SELECTIVE_DEBUG_LEVEL(AutotunePlugins),"First call to Create Scenarios\n");
        if( !automatic_eager.empty() ) {
            int mpi_msg_p2p_tot, mpi_msg_p2p_thr;
            int mpi_msg_freq_2k, mpi_msg_freq_4k, mpi_msg_freq_8k;
            int mpi_msg_freq_16k, mpi_msg_freq_32k, mpi_msg_freq_64k;
            int tot_msg_thr;

            map<int, list<MetaProperty> >                 preAnalysisPropertiesMap;
            map<int, list<MetaProperty> >::const_iterator propertyMapIter;
            list<MetaProperty>                            properties;
            list<MetaProperty>::iterator                  property;
            list<Region*>                                 regions;
            list<Region*>::iterator                       reg;

            preAnalysisPropertiesMap = pool_set->arp->getAllPreAnalysisProperties();

            for( propertyMapIter = preAnalysisPropertiesMap.begin(); propertyMapIter != preAnalysisPropertiesMap.end(); propertyMapIter++ ) {
                psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParameters PLugin: Getting %d properties from the PreAnalysisPropertiesMap \n",  preAnalysisPropertiesMap.size() );
                properties      = propertyMapIter->second;
                mpi_msg_p2p_tot = mpi_msg_p2p_thr = mpi_msg_freq_2k = mpi_msg_freq_4k = mpi_msg_freq_8k = mpi_msg_freq_16k = mpi_msg_freq_32k = mpi_msg_freq_64k = 0;
                tot_msg_thr     = 0;

                for( property = properties.begin(); property != properties.end(); property++ ) {
                    if( atoi( property->getId().c_str() ) == EAGERLIMITDEPENDENT ) {
                        mpi_msg_p2p_tot  += atoi( property->getExtraInfo().at( "mpi_msg_p2p_tot" ).c_str() );
                        mpi_msg_p2p_thr  += atoi( property->getExtraInfo().at( "mpi_msg_p2p_thr" ).c_str() );
                        mpi_msg_freq_2k  += atoi( property->getExtraInfo().at( "mpi_msg_freq_2k" ).c_str() );
                        mpi_msg_freq_4k  += atoi( property->getExtraInfo().at( "mpi_msg_freq_4k" ).c_str() );
                        mpi_msg_freq_8k  += atoi( property->getExtraInfo().at( "mpi_msg_freq_8k" ).c_str() );
                        mpi_msg_freq_16k += atoi( property->getExtraInfo().at( "mpi_msg_freq_16k" ).c_str() );
                        mpi_msg_freq_32k += atoi( property->getExtraInfo().at( "mpi_msg_freq_32k" ).c_str() );
                        mpi_msg_freq_64k += atoi( property->getExtraInfo().at( "mpi_msg_freq_64k" ).c_str() );
                    }
                }
                tot_msg_thr = mpi_msg_freq_2k + mpi_msg_freq_4k + mpi_msg_freq_8k + mpi_msg_freq_16k + mpi_msg_freq_32k + mpi_msg_freq_64k;

                if( ( mpi_msg_p2p_tot > 0 ) && ( mpi_msg_p2p_thr >= 0.3 * mpi_msg_p2p_tot ) ) {
                    //Tuning the eager limit is worthy because volume of bytes communicated within the eager limit range represent 30 or more of the total
                    unsigned lowlim  = 1024, uplim = 65536, step;
                    float    prop2k  = ( 1.0 * mpi_msg_freq_2k ) / tot_msg_thr;
                    float    prop4k  = ( 1.0 * mpi_msg_freq_4k ) / tot_msg_thr;
                    float    prop8k  = ( 1.0 * mpi_msg_freq_8k ) / tot_msg_thr;
                    float    prop16k = ( 1.0 * mpi_msg_freq_16k ) / tot_msg_thr;
                    float    prop32k = ( 1.0 * mpi_msg_freq_32k ) / tot_msg_thr;
                    float    prop64k = ( 1.0 * mpi_msg_freq_64k ) / tot_msg_thr;

                    if( prop2k > 0.5 ) {
                        uplim = 2048;
                        step  = 256;
                    }
                    else if( prop4k > 0.5 ) {
                        lowlim = 2048;
                        uplim  = 4096;
                        step   = 256;
                    }
                    else if( prop8k > 0.5 ) {
                        lowlim = 4096;
                        uplim  = 8192;
                        step   = 512;
                    }
                    else if( prop16k > 0.5 ) {
                        lowlim = 8192;
                        uplim  = 16384;
                        step   = 1024;
                    }
                    else if( prop32k > 0.5 ) {
                        lowlim = 16384;
                        uplim  = 32768;
                        step   = 1024;
                    }
                    else if( prop64k > 0.5 ) {
                        lowlim = 32768;
                        step   = 1024;
                    }

                    if( ( prop2k + prop4k ) > 0.6 ) {
                        uplim = 4096;
                        step  = 512;
                    }
                    if( ( prop4k + prop8k ) > 0.6 ) {
                        lowlim = 2048;
                        uplim  = 8192;
                        step   = 512;
                    }
                    if( ( prop8k + prop16k ) > 0.6 ) {
                        lowlim = 4096;
                        uplim  = 16384;
                        step   = 1024;
                    }
                    if( ( prop16k + prop32k ) > 0.6 ) {
                        lowlim = 8192;
                        uplim  = 32768;
                        step   = 1024;
                    }
                    if( ( prop32k + prop64k ) > 0.6 ) {
                        lowlim = 16384;
                        uplim  = 65536;
                        step   = 2048;
                    }

                    if( ( prop2k + prop4k + prop8k ) > 0.8 ) {
                        lowlim = 1024;
                        uplim  = 8192;
                        step   = 512;
                    }
                    if( ( prop4k + prop8k + prop16k ) > 0.8 ) {
                        lowlim = 2048;
                        uplim  = 16384;
                        step   = 1024;
                    }
                    if( ( prop8k + prop16k + prop32k ) > 0.8 ) {
                        lowlim = 4096;
                        uplim  = 32768;
                        step   = 2048;
                    }
                    if( ( prop16k + prop32k + prop64k ) > 0.8 ) {
                        lowlim = 8192;
                        uplim  = 65536;
                        step   = 2048;
                    }

                    if( ( prop2k + prop4k + prop8k + prop16k ) > 0.9 ) {
                        lowlim = 1024;
                        uplim  = 16384;
                        step   = 1024;
                    }
                    if( ( prop4k + prop8k + prop16k + prop32k ) > 0.9 ) {
                        lowlim = 2048;
                        uplim  = 32768;
                        step   = 2048;
                    }
                    if( ( prop8k + prop16k + prop32k + prop64k ) > 0.9 ) {
                        lowlim = 4096;
                        uplim  = 65536;
                        step   = 2048;
                    }

                    if( ( prop2k + prop4k + prop8k + prop16k + prop32k ) > 0.9 ) {
                        lowlim = 1024;
                        uplim  = 32768;
                        step   = 2048;
                    }
                    if( ( prop4k + prop8k + prop16k + prop32k + prop64k ) > 0.9 ) {
                        lowlim = 2048;
                        uplim  = 65536;
                        step   = 2048;
                    }

                    unsigned bufmemlow, bufmemup, bufmemstep;

                    int numprocs = context->getMPINumProcs();

                    bufmemlow  = 2 * numprocs * lowlim;
                    bufmemup   = 2 * numprocs * uplim;
                    bufmemstep = 2 * numprocs * step;

                    Region* pr = Application::instance().get_phase_region();

                    //Now we create the new tuning parameter

                    int tpid = tuningParameters.back()->getId() + 1;

                    TuningParameter* tp = new TuningParameter();

                    tp->setId( tpid );
                    tp->setName( automatic_eager.c_str() );
                    tp->setPluginType( MPI );
                    tp->setRuntimeActionType( TUNING_ACTION_NONE );
                    tp->setRange( lowlim, uplim, step );
                    Restriction* res = new Restriction();
                    res->setRegion( pr );
                    res->setRegionDefined( true );
                    tp->setRestriction( res );

                    tuningParameters.push_back( tp );

                    std::vector<string> values;
                    valuesMap.insert( make_pair( tp->getId(), values ) );

                    tp = new TuningParameter();
                    tp->setId( tpid + 1 );
                    tp->setName( automatic_buf_mem.c_str() );
                    tp->setPluginType( MPI );
                    tp->setRuntimeActionType( TUNING_ACTION_NONE );
                    tp->setRange( bufmemlow, bufmemup, bufmemstep );
                    res = new Restriction();
                    res->setRegion( pr );
                    res->setRegionDefined( true );
                    tp->setRestriction( res );

                    tuningParameters.push_back( tp );

                    valuesMap.insert( make_pair( tp->getId(), values ) );

                    if( active_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ) ) > 0 ) {
                        printf( "\nAll created tuning parameters\n=======================\n" );
                        for( int i = 0; i < tuningParameters.size(); i++ ) {
                            string str = tuningParameters[ i ]->toString();
                            cout << str <<  "=====================" << endl;
                        }
                    }
                }
            }
        }

        for( int i = 0; i < tuningParameters.size(); i++ ) {
            variantSpace.addTuningParameter( tuningParameters[ i ] );
        }
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: Create a SearchSpace from %d tuning parameters.\n", tuningParameters.size() );
        searchSpace.setVariantSpace( &variantSpace );

        searchSpace.addRegion( tuningParameters[ 0 ]->getRestriction()->getRegion() );
        searchAlgorithm->addSearchSpace( &searchSpace );
        pass = true;
        searchAlgorithm->createScenarios();
    }
    else {
        //psc_dbgmsg(PSC_SELECTIVE_DEBUG_LEVEL(AutotunePlugins),"Another call to Create Scenarios\n");

        searchAlgorithm->createScenarios();
    }
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: Scenarios created\n" );
}

/**
 * @brief Preparatory steps for the scenarios are done here.
 *
 * This method builds the set of flags that will be used for passing to the application
 * the values corresponding to a given scenario (obtained from the csp). In the case of the
 * supported MPI flavors (IBM MPI, OpenMPI, or Intel MPI) the parameters are passed in the
 * application command line using appflags, while, in any other case, they are set through
 * environment variables using exports.
 * Finally, the scenario is pushed into the prepared scenario pool (psp).
 *
 * @ingroup MPIParametersPlugin
 */
void MPIParametersPlugin::prepareScenarios( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to prepareScenarios()\n" );
    if( !pool_set->csp->empty() ) {
        Scenario*     scenario;
        ostringstream command;
        scenario = pool_set->csp->pop();
        const list<TuningSpecification*>* ts     = scenario->getTuningSpecifications();
        map<TuningParameter*, int>        values = ts->front()->getVariant()->getValue();

        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: recovered scenario from CSP and values [%d]\n", values.size() );
        //restartCommand.clear();
        appflags.clear();
        exports.clear();

        switch( MPIType ) {
        case IBM:
            for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                if( current_values.size() == 0 ) {
                    command << "-" << it->first->getName() << " " << it->second << " ";
                    appflags.append( command.str() );
                }
                else {
                    command << "-" << it->first->getName() << " " << current_values.at( it->second ) << " ";
                    appflags.append( command.str() );
                }
                command.str( std::string() );
                command.clear();
            }
            break;
        case OPENMPI:
            for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                if( current_values.size() == 0 ) {
                    command << "--mca " << it->first->getName() << " " << it->second << " ";
                    appflags.append( command.str() );
                }
                else {
                    command << "--mca " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                    appflags.append( command.str() );
                }
                command.str( std::string() );
                command.clear();
            }
            break;
        case INTEL:
            for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                if( current_values.size() == 0 ) {
                    command << "-genv " << it->first->getName() << " " << it->second << " ";
                    appflags.append( command.str() );
                }
                else {
                    command << "-genv " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                    appflags.append( command.str() );
                }
                command.str( std::string() );
                command.clear();
            }
            break;

        default:
            for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                if( current_values.size() == 0 ) {
                    command << "export " << it->first->getName() << "=" << it->second << "; ";
                    exports.append( command.str() );
                }
                else {
                    command << "export " << it->first->getName() << "=" << current_values.at( it->second ) << "; ";
                    exports.append( command.str() );
                }
                command.str( std::string() );
                command.clear();
            }
            break;
        }

        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: export environment string = %s\n", appflags.c_str() );
        pool_set->psp->push( scenario );
    }
}

/**
 * @brief Populate the Experiment Scenario Pool (esp) for the next experiment.
 *
 * This is the final step before the experiments are executed. Scenarios are simply moved from the
 * psp to the esp.
 *
 * @ingroup MPIParametersPlugin
 * @param numprocs Not used.
 * @param analysisRequired Set to false.
 * @param strategy No analysis strategy requested.

 */
void MPIParametersPlugin::defineExperiment( int               numprocs,
                                            bool&             analysisRequired,
                                            StrategyRequest** strategy ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to defineExperiment()\n" );

    if( !pool_set->psp->empty() ) {
        Scenario* scenario;
        scenario = pool_set->psp->pop();
        const list<TuningSpecification*>* ts = scenario->getTuningSpecifications();
        if( ts->size() != 1 ) {
            perror( "Can't currently handle multiple tuning specs\n" );
            throw 0;
        }

        list<unsigned int>*     ranks     = new list<unsigned int>;
        list<PropertyRequest*>* pr_in     = new list<PropertyRequest*>();
        list<int>*              prop_list = new list<int>;
        ranks->push_back( 0 );
        prop_list->push_back( EXECTIME );
        pr_in->push_back( new PropertyRequest( prop_list, ranks ) );
        scenario->setPropertyRequests( pr_in );
        //scenario->setTunedRegion(tuningParameters[0]->getRestriction()->getRegion());
        scenario->setTunedRegion( tunedRegion );


        pool_set->esp->push( scenario );
    }

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: Added 1 scenario in the experiment.\n" );
    0;
}

/**
 * @brief Return true if a restart of the application is required for the next experiment,
 * false otherwise.
 *
 *
 *
 * @ingroup MPIParametersPlugin
 * @param env Used for setting the MPI parameters through environment variables when
 * unsupported flavours are used.
 * @param numprocs not changed (the number of processes is not changed)
 * @param command for indicating the application command line invocation using the flags build at
 * @see prepareScenarios (for the supported MPI implementations)
 * @param is_instrumented
 *
 */
bool MPIParametersPlugin::restartRequired( std::string& env,
                                           int&         numprocs,
                                           std::string& command,
                                           bool&        is_instrumented ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to restartRequired()\n" );

    if( restartCommand.empty() ) {
        restartCommand.assign( command );
    }
    cout << "Restart command: " << restartCommand << endl;
    cout << "App flags: " << appflags << endl;
    switch( MPIType ) {
    case IBM:
        command.assign( restartCommand );
        command.append( " " );
        command.append( appflags );
        break;
    case OPENMPI:
        command.assign( appflags );
        command.append( " " );
        command.append( restartCommand );
        break;
    case INTEL:
        command.assign( appflags );
        command.append( " " );
        command.append( restartCommand );
        break;
    default:
        env.assign( exports );
        break;
    }
//  exports.clear();
    appflags.clear();
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: outgoing env: %s\n", env.c_str() );
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: outgoing command: %s\n", command.c_str() );
    is_instrumented = !context->applUninstrumented();
    return true;
}

/**
 * @brief Returns whether the search algorithm is finished.
 * Return true if if the current search iteration is finished (by either the plugin itself or as
 * a result of the search algorithm), false otherwise.
 *
 * @ingroup MPIParameters Plugin
 */
bool MPIParametersPlugin::searchFinished( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to searchFinished()\n" );

    return searchAlgorithm->searchFinished();
}

/**
 * @brief Final operation of a tuning step.
 *
 * If any post-processing is necessary before entering the next tuning iteration, it is to be
 * done here.
 *
 * @ingroup MPIParametersPlugin
 */
void MPIParametersPlugin::finishTuningStep( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to processResults()\n" );
}

/**
 * @brief Returns true if the plugin finished the tuning process, false otherwise.
 * @ingroup MPIParametersPlugin
 * @return True
 */
bool MPIParametersPlugin::tuningFinished( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to tuningFinished()\n" );

    return true;
}

/**
 * @brief Prints to the screen (and to a file, where necessary) the tuning advice.
 *
 * Prints the best combination of the flags, as well as the results for all the tested combinations.
 *
 * @ingroup MPIParametersPlugin
 */
Advice* MPIParametersPlugin::getAdvice( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getAdvice()\n" );
    if( searchAlgorithm == NULL ) {
        throw PTF_PLUGIN_ERROR( NULL_REFERENCE );
    }

// Storing scenario execution time into the file
/******************************************************************************/
    ofstream          outfile;
    std::stringstream filename;
    const char*       separator = "";
//  filename << "results_MPIParameters" << (int) getpid() << ".csv";
//  filename << "results_MPIParameters.csv";
    outfile.open( "results_MPIParameters.csv", ios_base::app );
/******************************************************************************/

    cout << "\n\nFound best scenario: " << searchAlgorithm->getOptimum() << endl;
    cout << "Parameter combination: \n";
    const list<TuningSpecification*>* ts = pool_set->fsp->getTuningSpecificationByScenarioID(
        searchAlgorithm->getOptimum() );
    map<TuningParameter*, int> values = ts->front()->getVariant()->getValue();
    string                     appflags, exports;
    ostringstream              command;

    switch( MPIType ) {
    case IBM:
        for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
            vector<string> current_values = valuesMap.find( it->first->getId() )->second;
            if( current_values.size() == 0 ) {
                command << "-" << it->first->getName() << " " << it->second << " ";
                appflags.append( command.str() );
            }
            else {
                command << "-" << it->first->getName() << " " << current_values.at( it->second ) << " ";
                appflags.append( command.str() );
            }
            command.str( std::string() );
            command.clear();
        }
        cout << appflags;
        break;
    case OPENMPI:
        for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
            vector<string> current_values = valuesMap.find( it->first->getId() )->second;
            if( current_values.size() == 0 ) {
                command << "--mca " << it->first->getName() << " " << it->second << " ";
                appflags.append( command.str() );
            }
            else {
                command << "--mca " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                appflags.append( command.str() );
            }
            command.str( std::string() );
            command.clear();
        }
        cout << appflags;
        break;
    case INTEL:
        for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
            vector<string> current_values = valuesMap.find( it->first->getId() )->second;
            if( current_values.size() == 0 ) {
                command << "-genv " << it->first->getName() << " " << it->second << " ";
                appflags.append( command.str() );
            }
            else {
                command << "-genv " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                appflags.append( command.str() );
            }
            command.str( std::string() );
            command.clear();
        }
        cout << appflags;
    default:
        for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
            vector<string> current_values = valuesMap.find( it->first->getId() )->second;
            if( current_values.size() == 0 ) {
                command << "export " << it->first->getName() << "=" << it->second << "; ";
                exports.append( command.str() );
            }
            else {
                command << "export " << it->first->getName() << "=" << current_values.at( it->second ) << "; ";
                exports.append( command.str() );
            }
            command.str( std::string() );
            command.clear();
        }
        cout << appflags;
        break;
    }
    appflags.clear();

    cout << "\n\nAll Results:\n";
    cout << "Scenario  |   Runtime    | Flags\n";
    for( int scenario_id = 0; scenario_id < pool_set->fsp->size(); scenario_id++ ) {
        if( !pool_set->srp->getScenarioResultsByID( scenario_id ).empty() ) {
            printf( "%-10d|", scenario_id );
            list<MetaProperty> properties = pool_set->srp->getScenarioResultsByID( scenario_id );
            for( std::list<MetaProperty>::iterator iterator = properties.begin(), end = properties.end(); iterator != end;
                 ++iterator ) {
                printf( "\t%f | ", iterator->getSeverity() );
                outfile << separator << iterator->getSeverity();
                separator = "; ";
            }

            const list<TuningSpecification*>* ts     = pool_set->fsp->getTuningSpecificationByScenarioID( scenario_id );
            map<TuningParameter*, int>        values = ts->front()->getVariant()->getValue();

            switch( MPIType ) {
            case IBM:
                for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                    vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                    if( current_values.size() == 0 ) {
                        command << "-" << it->first->getName() << " " << it->second << " ";
                        appflags.append( command.str() );
                    }
                    else {
                        command << "-" << it->first->getName() << " " << current_values.at( it->second ) << " ";
                        appflags.append( command.str() );
                    }
                    command.str( std::string() );
                    command.clear();
                }
                cout << appflags;
                break;
            case OPENMPI:

                for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                    vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                    if( current_values.size() == 0 ) {
                        command << "--mca " << it->first->getName() << " " << it->second << " ";
                        appflags.append( command.str() );
                    }
                    else {
                        command << "--mca " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                        appflags.append( command.str() );
                    }
                    command.str( std::string() );
                    command.clear();
                }
                cout << appflags;
                break;
            case INTEL:
                for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                    vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                    if( current_values.size() == 0 ) {
                        command << "-genv " << it->first->getName() << " " << it->second << " ";
                        appflags.append( command.str() );
                    }
                    else {
                        command << "-genv " << it->first->getName() << " " << current_values.at( it->second ) << " ";
                        appflags.append( command.str() );
                    }
                    command.str( std::string() );
                    command.clear();
                }
                cout << appflags;
                break;
            default:
                for( map<TuningParameter*, int>::const_iterator it = values.begin(); it != values.end(); ++it ) {
                    vector<string> current_values = valuesMap.find( it->first->getId() )->second;
                    if( current_values.size() == 0 ) {
                        command << "export " << it->first->getName() << "=" << it->second << "; ";
                        exports.append( command.str() );
                    }
                    else {
                        command << "export " << it->first->getName() << "=" << current_values.at( it->second ) << "; ";
                        exports.append( command.str() );
                    }
                    command.str( std::string() );
                    command.clear();
                }
                cout << appflags;
                break;
            }
            printf( "\n" );
            appflags.clear();
        }
    }

    outfile << endl;
// TODO: Find out why fsp has no results and the advice is therefore missing results
//  cout << pool_set->fsp->toString(1, "\t") << endl;

    return new Advice( getName(), ( *pool_set->fsp->getScenarios() )[ searchAlgorithm->getOptimum() ],
                       searchAlgorithm->getSearchPath(), "Time", pool_set->fsp->getScenarios() );
}

/**
 * @brief Finalizes the plugin.
 * @ingroup MPIParametersPlugin
 *
 */
void MPIParametersPlugin::finalize() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to finalize()\n" );

    terminate();
}

/**
 * @brief Terminate the plugin due to error.
 *
 * Safely remove any allocated memory, objects, file descriptors, etc. This method should
 * be able to be executed safely at any point.
 *
 * @ingroup MPIParametersPlugin
 */
void MPIParametersPlugin::terminate() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to terminate()\n" );
    if( searchAlgorithm ) {
        searchAlgorithm->finalize();
        delete searchAlgorithm;
    }
    context->unloadSearchAlgorithms();
}

/**
 * The following are C functions that belong to the Plugin Management Interface, and not to the
 * IPlugin class. Currently the interface includes information and instance creation functions.
 *
 * These are defined in PluginManagement.h .
 */

/**
 * @brief Returns Plugin's instance.

 * Returns an instance of this particular plugin implementation.
 * Typically, a simple return with new is enough. For example:
 *
 * @ingroup MPIParametersPlugin
 * @return new MPIParametersPlugin();
 */
IPlugin* getPluginInstance( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getPluginInstance()\n" );

    return new MPIParametersPlugin();
}

/**
 * @brief Returns the major plugin interface version used by this plugin (example: the 1 in 1.0).
 * @ingroup MPIParametersPlugin
 * @return 1
 */
int getVersionMajor( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getInterfaceVersionMajor()\n" );

    return MPIPARAMETERS_VERSION_MAJOR;
}

/**
 * @brief Returns the minor plugin interface version used by this plugin (example: the 0 in 1.0).
 * @ingroup MPIParametersPlugin
 * @return 0
 */
int getVersionMinor( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getInterfaceVersionMinor()\n" );

    return MPIPARAMETERS_VERSION_MINOR;
}

/**
 * @brief Returns a string with the name of the plugin.
 * @ingroup MPIParametersPlugin
 *
 * @return A string with the name of the plugin.
 */
string getName( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getName()\n" );

    return "MPI Parameter Plugin";
}

/**
 * @brief Returns a string with a short description of the plugin.
 * @ingroup MPIParametersPlugin
 *
 * @return A string with a short description of the plugin.
 */
string getShortSummary( void ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( AutotunePlugins ), "MPIParametersPlugin: call to getShortSummary()\n" );

    return "Finds optimal combination of a subset of MPI library parameters based on a goal metric (execution time)";
}
