/**
   @file    AltixPBSStarter.cc
   @ingroup Frontend
   @brief   PBS starter for Altix
   @author  Michael Gerndt
   @verbatim
    Revision:       $Revision$
    Revision date:  $Date$
    Committed by:   $Author$

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2011, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include "AltixPBSStarter.h"
#include <string>
#include <cstring>
#include <sstream>
#include <unistd.h>
#include <sys/types.h>  // Process id type return by fork(): pid_t

#include "frontend.h"
#include "BinStorage.h"
#include "timing.h"
#include "regxx.h"
#include "psc_errmsg.h"
#include "psc_agent.h"
#include "selective_debug.h"

#pragma GCC diagnostic ignored "-Wwrite-strings"

using namespace std;



AltixPBSStarter::AltixPBSStarter() : Starter() {
    // TODO Auto-generated constructor stub
    stg = new BinStorage( 19, 512 );
    constructPBSnodelist();
    idmap_len = 0;
    for( int i = 0; i < 19; i++ ) {
        idmap_h_num[ i ] = 0;
    }
}

AltixPBSStarter::~AltixPBSStarter() {
    // TODO Auto-generated destructor stub
}

void AltixPBSStarter::constructPBSnodelist() {
    // read out the nodefile
    FILE* pnfp               = fopen( getenv( "PBS_NODEFILE" ), "rt" );
    int   relativeNodeNumber = 0;
    char  nfline[ 512 ];
    pbsnodelist_size = 0;
    while( fscanf( pnfp, "%s\n", nfline ) != EOF ) {
        //psc_dbgmsg( 1,"PBS_NODEFILE line:%s:\n", nfline );
        bool found = false;
        //search if node already exists in list then increase number of cpus for that node
        for( int ct = 0; ct < pbsnodelist_size; ct++ ) {
            if( !strcmp( nfline, pbsnodelist_string[ ct ].c_str() ) ) {
                // node already in the list
                found                     = true;
                pbsnodelist_amount[ ct ] += 1;
                break;
            }
        }
        if( !found ) {
            // new node, extend list
            pbsnodelist_string[ pbsnodelist_size ]   = nfline;
            pbsnodelist_amount[ pbsnodelist_size ]   = 1;
            pbsnodelist_firstidx[ pbsnodelist_size ] = relativeNodeNumber;
            pbsnodelist_size++;
        }
        relativeNodeNumber++;
    }

    for( int ct = 0; ct < pbsnodelist_size; ct++ ) {
        pbsnodelist_amount[ pbsnodelist_size ] = 1;
        psc_dbgmsg( 1, "node >%s< with %i cpus\n", pbsnodelist_string[ ct ].c_str(), pbsnodelist_amount[ ct ] );
        pbsnodelist_amount[ ct ] = stg->read_cpuset( ct, pbsnodelist_firstidx[ ct ], getpid() + ct );
    }
}

bool AltixPBSStarter::runApplication() {
    stringstream     appstart_cmd;
    char*            senvptr_old = NULL;
    int              clusterSize = 128;
    RegistryService* regsrv      = fe->get_registry();
    pid_t            pid;
    int              child_status;

    if( opts.has_maxcluster ) {
        clusterSize = atoi( opts.maxcluster_string );
    }

    int   ct;
    pid_t procid = getpid();

    int mindfp = 1;
    if( threads > 0 ) {
        mindfp = threads;
    }

    int totalprocs = 0;
    // get the total number of available processors
    for( ct = 0; ct < pbsnodelist_size; ct++ ) {
        pbsnodelist_distributed[ ct ] = 0;
        if( pbsnodelist_amount[ ct ] > mindfp ) {
            totalprocs += pbsnodelist_amount[ ct ];
        }
    }

    float dfrac = ( float )processes * mindfp / totalprocs;
    psc_dbgmsg( 5, " %f percent of %d CPUs required for application \n", dfrac * 100, totalprocs );

    if( dfrac > 1.0 ) {
        psc_errmsg( "not enough processors allocated!\n" );
        exit( 1 );
    }

    int totaldistributed = 0;
    // calculate the number of processors to be used by the
    // application on each node
    for( ct = 0; ct < pbsnodelist_size; ct++ ) {
        if( pbsnodelist_amount[ ct ] < mindfp + 1 ) {
            continue;
        }
        int procspernode = ( int )( pbsnodelist_amount[ ct ] * dfrac );

        psc_dbgmsg( 1, "usable cpus on node %i: %i/%i/%i\n",
                    ct, procspernode - ( procspernode % mindfp ), procspernode, pbsnodelist_amount[ ct ] );
        procspernode                  = procspernode - ( procspernode % mindfp );
        pbsnodelist_distributed[ ct ] = procspernode;
        totaldistributed             += procspernode;
    }


    // another iteration that cares about small-sized
    // nodes that might not have been taken into account yet
    if( totaldistributed < processes ) {
        for( ct = 0; ct < pbsnodelist_size; ct++ ) {
            if( ( pbsnodelist_distributed[ ct ] == 0 ) && ( pbsnodelist_amount[ ct ] > mindfp ) ) {
                pbsnodelist_distributed[ ct ] = mindfp;
                totaldistributed             += mindfp;
                if( totaldistributed >= processes * mindfp ) {
                    break;
                }
            }
        }
    }


    // last iteration that allocates possibly missing cpus
    // (when the fractional distribution did not make up the
    // total number of cpus to distribute)
    while( totaldistributed < processes * mindfp ) {
        bool procadded = false;
        for( ct = 0; ct < pbsnodelist_size; ct++ ) {
            if( pbsnodelist_amount[ ct ] < mindfp + 1 ) {
                continue;
            }
            if( pbsnodelist_distributed[ ct ] + mindfp < pbsnodelist_amount[ ct ] ) {
                pbsnodelist_distributed[ ct ] += mindfp;
                totaldistributed              += mindfp;
                if( totaldistributed >= processes * mindfp ) {
                    break;
                }
                procadded = true;
            }
        }
        if( !procadded ) {
            break;
        }
    }

    for( ct = 0; ct < pbsnodelist_size; ct++ ) {
        psc_dbgmsg( 5, "processes/threads to be placed on node %d: %d\n", ct, pbsnodelist_distributed[ ct ] );
    }

    int max_cpus_per_bin = clusterSize;
    if( threads > 1 ) {
        // hybrid application, special treatment
        int min_cpus_per_bin  = threads;
        int max_procs_per_bin = max_cpus_per_bin / ( min_cpus_per_bin ) + 1;
        max_cpus_per_bin = max_procs_per_bin * min_cpus_per_bin;
        psc_dbgmsg( 5, "generating clusters with %d mpi processes per cluster and %d cpus per cluster (processes*threads)\n",
                    max_procs_per_bin, max_cpus_per_bin );
    }


    //MG: computation of clusters/bins and thus number of node agents
    for( int sct = 0; sct < 1; sct++ ) {
        //stg->clear_data();

        // information retrieval and clustering for all nodes
        for( int lct = 0; lct < pbsnodelist_size; lct++ ) {
            //psc_dbgmsg( 1,"gathering information about node %d\n", lct );

            //stg->read_cpuset( lct, pbsnodelist_firstidx[lct], procid + lct );
            //stg->read_cpuset( lct, lct, procid + lct );

            stg->read_topology( lct, pbsnodelist_firstidx[ lct ], procid + lct );
            //stg->read_topology( lct, lct, procid );

            psc_dbgmsg( 5, "generating clusters controlled by a single analysis agent for node %d\n", lct );
            // partition the allocated cpus into logical clusters
            stg->generate_bins( lct, max_cpus_per_bin );

            // see what bins will be actually needed
            stg->activate_bins( lct, pbsnodelist_distributed[ lct ], threads );
        }

        psc_dbgmsg( 5, "start cluster optimization\n" );

        // optimize logical clusters
        for( ct = 0; ct < pbsnodelist_size; ct++ ) {
            stg->optimize_bins( ct );
//            stg->optimize_bins_siman( ct );
//            stg->optimize_bins_rand( ct );
        }

        psc_dbgmsg( 5, "finished cluster optimization\n" );

        // debug output of the current bin distribution
//        for ( ct = 0; ct < pbsnodelist_size; ct++ )
//        {
//            stg->output_current_bins( ct );
//        }
    }
    char mri_debug[ 1000 ] = "";

    if( active_dbgLevel( 1004 ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( 1004 ) );
    }
    if( active_dbgLevel( 1011 ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( 1011 ) );
    }

    appstart_cmd << "export PSC_MRI_DEBUG=" << mri_debug << ";";

    psc_dbgmsg( 5, "generating placement command...\n" );
    bool first = true;
    for( ct = 0; ct < pbsnodelist_size; ct++ ) {
        if( pbsnodelist_distributed[ ct ] / mindfp > 0 ) {
            if( first ) {
                // generate mpirun-command
                appstart_cmd << "export PSC_REGISTRY=" << regsrv->get_reghost() << ":" << regsrv->get_regport() << "; export OMP_NUM_THREADS=" << threads << "; export PSC_APPNAME=" << fe->get_appname() << ";mpirun -v "
                             << pbsnodelist_string[ ct ].c_str()
                             << " -np "
                             << pbsnodelist_distributed[ ct ] / mindfp
                             << " /usr/pbs/default/bin/pbs_attach -j $PBS_JOBID dplace -p ";
                first = false;
            }
            else {
                appstart_cmd << " : "
                             << pbsnodelist_string[ ct ].c_str()
                             << " -np "
                             << pbsnodelist_distributed[ ct ] / mindfp
                             << " /usr/pbs/default/bin/pbs_attach -j $PBS_JOBID dplace -p ";
            }


            char tmpname[ 512 ];
            sprintf( tmpname, "%s/pscmap3%i", getenv( "OPT_TMP" ), getpid() + ct );
            stg->create_placement_file( ct, tmpname, threads, pbsnodelist_distributed[ ct ] );
            appstart_cmd << tmpname << " " << opts.app_run_string;
        }
    }

    char senvbuf[ 1000 ];
    if( threads > 0 ) {
        // set number of OpenMP-Threads of the application
        sprintf( senvbuf, "MPI_OMP_NUM_THREADS=%i", threads );
        char tmpbuf[ 512 ];
        for( int nct = 1; nct < pbsnodelist_size; nct++ ) {
            sprintf( tmpbuf, ",%i", threads );
            strcat( senvbuf, tmpbuf );
        }
        psc_dbgmsg( 5, "OpenMP environment string: >%s<\n", senvbuf );
        putenv( senvbuf );
        putenv( "MPI_OPENMP_INTEROP=yes" );
    }

    pid = fork();
    if( pid == 0 ) {
        char appbuf[ 512 ];
        sprintf( appbuf, "PSC_APPNAME=%s", fe->get_appname() );
        psc_dbgmsg( 1, "APP string: >%s<\n", fe->get_appname() );
        putenv( appbuf );

        string help = appstart_cmd.str();
        psc_dbgmsg( 1, "starting application: >%s<\n", help.c_str() );
        int retVal = system( help.c_str() );
        exit( retVal );
    }
    else if( pid < 0 ) {
        psc_errmsg( "Error forking child process!\n" );
        return false;
    }

    if( senvptr_old == NULL ) {
//                putenv("OMP_NUM_THREADS=0");
        if( processes > 0 ) {
            putenv( "MPI_OPENMP_INTEROP=no" );
        }
    }
    else {
        putenv( senvptr_old );
    }
    //exit(0);

    EntryData query;

    bool startup_finished       = false;
    int  num_startup_components = processes;

    psc_dbgmsg( 5, "waiting for startup of %d components...\n", num_startup_components );

    if( num_startup_components > 1024 ) {
        sleep( 20 );
    }
    else if( num_startup_components > 256 ) {
        sleep( 10 );
    }

    //compute mapping of ranks to registry ids
    int delay_interval_dt = 3;
    sleep( 30 );

    std::list< EntryData > rresult;
    while( !startup_finished &&     !fe->timed_out() ) {
        sleep( 3 );
        query.app  = fe->get_appname();
        query.comp = "MRIMONITOR";
        query.tag  = "none";

        rresult.clear();
        if( regsrv->query_entries( rresult, query, false ) == -1 ) {
            psc_errmsg( "Error querying registry for application %s\n", fe->get_appname() );
            exit( 1 );
        }

        psc_dbgmsg( 1, "components started up: %d (of %d)\n", rresult.size(), num_startup_components );
        startup_finished = ( rresult.size() >= num_startup_components );
        if( rresult.size() > num_startup_components ) {
            psc_errmsg( "Error in startup of application processes. %d processes found but only %d expected.\n", rresult.size(), num_startup_components );
            exit( 1 );
        }

        // Child finished but components not started -> process was terminated or error occurred
        if( ( waitpid( pid, &child_status, WNOHANG ) == pid ) && ( !startup_finished ) ) {
            return false;
        }

        if( !startup_finished ) {
            sleep( delay_interval_dt );                          //sss
        }
    }

    if( !startup_finished ) {
        psc_errmsg( "Error in startup of application processes. Startup timed out.\n" );
        exit( 1 );
    }

    for( int pr = 0; pr < processes; pr++ ) {
        idmap_f[ pr ] = -1;
        idmap_t[ pr ] = -1;
    }

    std::list< EntryData >::iterator entryit;
    for( entryit = rresult.begin(); entryit != rresult.end(); entryit++ ) {
        EntryData entry;
        entry                    = *entryit;
        idmap_t[ entry.pid - 1 ] = ( *entryit ).id;

//        std::list<std::string>::iterator stringit;
//        for ( stringit = entry.strings.begin(); stringit != entry.strings.end(); stringit++ )
//        {
//            int maxrank;
//            int myrank;
//            int mycpu;
//            if( sscanf( stringit->c_str(), "PERISCOPE processrank [%d:%d:%d]",
//                &maxrank, &myrank, &mycpu ) > 0 )
//            {
//                //psc_dbgmsg( 2, "Entry %d: node %s, rank %d, mycpu %d, port %d\n", ( *entryit ).id, ( *entryit ).node.c_str(), myrank, mycpu, ( *entryit ).port );
//                      idmap_t[ myrank ] = ( *entryit ).id;
//            }
//        }
    }

    for( int i = 0; i < processes; i++ ) {
        if( idmap_t[ i ] == -1 ) {
            psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n", i );
            abort();
        }
    }

    return true;
}

bool AltixPBSStarter::rerunApplication() {
    RegistryService*       regsrv = fe->get_registry();
    std::list< EntryData > rresult;
    int                    mindfp = 1, ct;
    stringstream           appstart_cmd;
    char*                  senvptr_old = NULL;
    pid_t                  pid;
    int                    child_status;

    for( int rank = 0; rank < processes; rank++ ) {
        idmap_f[ rank ] = idmap_t[ rank ];
        idmap_t[ rank ] = -1;
    }

    /**
     * Set time-out to the maximum value and call reinstrumetation procedure on restart
     */
    fe->set_global_timeout( 429496729 );
    instrumentRequiredRegions();

    char mri_debug[ 1000 ] = "";

    if( active_dbgLevel( 1004 ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( 1004 ) );
    }

    if( active_dbgLevel( 1011 ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( 1011 ) );
    }

    appstart_cmd << "export PSC_MRI_DEBUG=" << mri_debug << ";";

    psc_dbgmsg( 5, "generating placement command...\n" );
    bool first = true;
    if( threads > 0 ) {
        mindfp = threads;
    }
    for( ct = 0; ct < pbsnodelist_size; ct++ ) {
        if( pbsnodelist_distributed[ ct ] / mindfp > 0 ) {
            if( first ) {
                // generate mpirun-command
                appstart_cmd << "export PSC_REGISTRY=" << regsrv->get_reghost() << ":" << regsrv->get_regport() << "; export OMP_NUM_THREADS=" << threads << "; export PSC_APPNAME=" << fe->get_appname() << ";mpirun -v "
                             << pbsnodelist_string[ ct ].c_str()
                             << " -np "
                             << pbsnodelist_distributed[ ct ] / mindfp
                             << " /usr/pbs/default/bin/pbs_attach -j $PBS_JOBID dplace -p ";
                first = false;
            }
            else {
                appstart_cmd << " : "
                             << pbsnodelist_string[ ct ].c_str()
                             << " -np "
                             << pbsnodelist_distributed[ ct ] / mindfp
                             << " /usr/pbs/default/bin/pbs_attach -j $PBS_JOBID dplace -p ";
            }


            char tmpname[ 512 ];
            sprintf( tmpname, "%s/pscmap3%i", getenv( "OPT_TMP" ), getpid() + ct );
            stg->create_placement_file( ct, tmpname, threads, pbsnodelist_distributed[ ct ] );
            appstart_cmd << tmpname << " " << opts.app_run_string;
        }
    }

    char senvbuf[ 1000 ];
    if( threads > 0 ) {
        // set number of OpenMP-Threads of the application
        sprintf( senvbuf, "MPI_OMP_NUM_THREADS=%i", threads );
        char tmpbuf[ 512 ];
        for( int nct = 1; nct < pbsnodelist_size; nct++ ) {
            sprintf( tmpbuf, ",%i", threads );
            strcat( senvbuf, tmpbuf );
        }
        psc_dbgmsg( 5, "OpenMP environment string: >%s<\n", senvbuf );
        putenv( senvbuf );
        putenv( "MPI_OPENMP_INTEROP=yes" );
    }

    pid = fork();
    if( pid == 0 ) {
        char appbuf[ 512 ];
        sprintf( appbuf, "PSC_APPNAME=%s", fe->get_appname() );
        psc_dbgmsg( 1, "APP string: >%s<\n", opts.app_run_string );
        putenv( appbuf );

        string help = appstart_cmd.str();
        psc_dbgmsg( 1, "starting application: >%s<\n", help.c_str() );
        int retVal = system( help.c_str() );
        _exit( retVal );
    }
    else if( pid < 0 ) {
        psc_errmsg( "Error forking child process!\n" );
        return false;
    }


    if( senvptr_old == NULL ) {
//        putenv( "OMP_NUM_THREADS=0" );
        if( processes > 0 ) {
            putenv( "MPI_OPENMP_INTEROP=no" );
        }
    }
    else {
        putenv( senvptr_old );
    }

    EntryData query;

    bool startup_finished  = false;
    int  delay_interval_dt = 2;


    if( processes > 16 ) {
        delay_interval_dt = 4;
    }
    else if( processes > 64 ) {
        delay_interval_dt = 8;
    }
    else if( processes > 1024 ) {
        delay_interval_dt = 15;
    }

    psc_dbgmsg( 5, "waiting for startup of %d components...\n", processes );
    while( !startup_finished ) {
        sleep( delay_interval_dt );
        query.app  = fe->get_appname();
        query.comp = "MRIMONITOR";
        query.tag  = "none";

        if( regsrv->query_entries( rresult, query, false ) == -1 ) {
            psc_errmsg( "Error querying registry for application\n" );
            exit( 1 );
        }

        psc_dbgmsg( 1, "components started up: %d (of %d)\n", rresult.size(), processes );
        startup_finished = ( rresult.size() >= processes );

        // Child finished but components not started -> process was terminated or error occured
        if( ( waitpid( pid, &child_status, WNOHANG ) == pid ) && ( !startup_finished ) ) {
            return false;
        }
    }     //waiting for all processes to register

    std::list< EntryData >::iterator entryit;
    for( entryit = rresult.begin(); entryit != rresult.end(); entryit++ ) {
        EntryData entry;
        entry                    = *entryit;
        idmap_t[ entry.pid - 1 ] = ( *entryit ).id;

//        std::list<std::string>::iterator stringit;
//        for ( stringit = entry.strings.begin(); stringit != entry.strings.end(); stringit++ )
//        {
//            int maxrank;
//            int myrank;
//            int mycpu;
//            if( sscanf( stringit->c_str(), "PERISCOPE processrank [%d:%d:%d]",
//                &maxrank, &myrank, &mycpu ) > 0 )
//            {
//                //psc_dbgmsg( 2, "Entry %d: node %s, rank %d, mycpu %d, port %d\n", ( *entryit ).id, ( *entryit ).node.c_str(), myrank, mycpu, ( *entryit ).port );
//                      idmap_t[ myrank ] = ( *entryit ).id;
//            }
//        }
    }

    for( int i = 0; i < processes; i++ ) {
        if( idmap_t[ i ] == -1 ) {
            psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n", i );
            abort();
        }
    }

    //send mapping info to child agents
    std::map< std::string, AgentInfo >*          child_agents = fe->get_child_agents();
    std::map< std::string, AgentInfo >::iterator it;
    for( it = ( *child_agents ).begin(); it != ( *child_agents ).end(); it++ ) {
        AgentInfo& ai = it->second;
        if( ai.status != AgentInfo::CONNECTED ) {
            if( fe->connect_to_child( &ai ) == -1 ) {
                psc_errmsg( "Error connecting to child at %s:%d\n",
                            ai.hostname.c_str(), ai.port );
            }
            else {
                ai.handler->reinit( processes, idmap_f, idmap_t );
            }
        }
    }


//
//
//    bool                                              foundany = false;
//    std::list< int >::iterator                        idit;
//    std::map< std::string, std::list<int> >           apphosts;
//    std::map< std::string, std::list<int> >::iterator apphostit;
//    RegistryService*                                  regsrv  = fe->get_registry();
//
//    for ( int ct = 0; ct < 8192; ct++ )
//    {
//        idmap_f[ ct ] = idmap_t[ ct ];
//        idmap_t[ ct ] = -1;
//    }
//
//
//    for( apphostit = apphosts.begin(); apphostit != apphosts.end(); apphostit++ )
//    {
////        strcat( masteragent_apphosts, ( *apphostit ).first.c_str() );
//
//        for( idit = ( *apphostit ).second.begin(); idit != ( *apphostit ).second.end(); idit++ )
//        {
//            EntryData entry;
//
//            if( regsrv->get_entry( entry, *idit, true ) ==-1 )
//            {
//                psc_errmsg( "Error querying data for entry %d\n", *idit );
//                continue;
//            }
//
//            std::list<std::string>::iterator it;
//            int                              num = 1;
//            for( it = entry.strings.begin(); it != entry.strings.end(); it++ )
//            {
////                psc_dbgmsg( 3, "String %d: %s\n", num++, it->c_str() );
//                int maxrank;
//                int myrank;
//                int mycpu;
//                if( sscanf( it->c_str(), "PERISCOPE processrank [%d:%d:%d]",
//                    &maxrank, &myrank, &mycpu ) > 0 )
//                {
//                    foundany = true;
//                    if ( myrank > idmap_len )
//                    {
//                        idmap_len = myrank;
//                    }
//                    if ( maxrank > idmap_len )
//                    {
//                        idmap_len = maxrank;
//                    }
//                    idmap_t[ myrank ] = *idit;
//                }
//            }
//        }
//    }
//
//    if ( !foundany )
//    {
//        for( apphostit = apphosts.begin(); apphostit != apphosts.end(); apphostit++ )
//        {
//            for( idit = ( *apphostit ).second.begin(); idit != ( *apphostit ).second.end(); idit++ )
//            {
//                EntryData entry;
//
//                if ( regsrv->get_entry( entry, *idit, true ) == -1 )
//                {
//                    psc_errmsg( "Error querying data for entry %d\n", *idit );
//                    continue;
//                }
//                idmap_len    = 1;
//                idmap_t[ 0 ] = *idit;
//            }
//        }
//    }
//
//    psc_dbgmsg( 1, "read %d IDs\n", idmap_len );
//    for (int ct = 0; ct < idmap_len; ct++ )
//    {
//        psc_dbgmsg( 1, "pos %d ID %d -> %d\n", ct, idmap_f[ ct ], idmap_t[ ct ] );
//    }
//
//    //send mapping info to child agents
//    std::map< std::string, AgentInfo >*          child_agents = fe->get_child_agents();
//    std::map< std::string, AgentInfo >::iterator it;
//    for( it = ( *child_agents ).begin(); it != ( *child_agents ).end(); it++ )
//    {
//        AgentInfo &ai = it->second;
//        if ( ai.status!=AgentInfo::CONNECTED
//        {
//            if ( fe->connect_to_child( &ai ) == -1 )
//            {
//                psc_errmsg( "Error connecting to child at %s:%d\n",
//                            ai.hostname.c_str(), ai.port );
//            }
//            else
//            {
//                ai.handler->reinit( idmap_len, idmap_f, idmap_t );
//            }
//        }
//    }

    return true;
}

PBSLevelInfo* AltixPBSStarter::computeAgentHierarchy() {
    return stg->BinsToAgentHierarchy( pbsnodelist_size, pbsnodelist_string, threads, processes, idmap_t, pbsnodelist_distributed );
}

void AltixPBSStarter::printAgentHierarchy( PBSLevelInfo* levels ) {
    PBSLevelInfo* level;

    level = levels;
    while( level != NULL ) {
        PBSAgentDetails* agent;
        agent = level->agents;
        psc_dbgmsg( 1, "Agents on this level: %d\n", level->numberOfAgents );
        while( agent != NULL ) {
            psc_dbgmsg( 1, "%s Port: %s, Host: %s, Rel. CPU: %s, Parent: %s, ApplIds: %s, Children: %s\n",
                        agent->tag, agent->port, agent->host, agent->relCPU, agent->parent, agent->applIds, agent->children );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }
}

int AltixPBSStarter::getHostNumber( char* host ) {
    for( int node = 0; node < pbsnodelist_size; node++ ) {
        if( !strcmp( host, pbsnodelist_string[ node ].c_str() ) ) {
            return node;
        }
    }
    psc_errmsg( "host %s not found!\n", host );
    exit( 1 );
}

void AltixPBSStarter::runAnalysisAgent( PBSAgentDetails* agent ) {
    char command_string[ 10000 ];
    char buf[ 1000 ];
    int  hostNumber = getHostNumber( agent->host );

    sprintf( command_string, "/usr/pbs/default/bin/pbsdsh -n %i -- /usr/pbs/default/bin/pbs_attach -j $PBS_JOBID /bin/bash -c \"cd %s; "
             "export USER=$PBS_O_LOGNAME; export LOGNAME=$USER; export PBS_NODEFILE=/var/spool/PBS/aux/$PBS_JOBID; . /etc/profile; ", pbsnodelist_firstidx[ hostNumber ], get_current_dir_name() );

//
//    char *delay = getenv( "PSC_STRATEGY_DELAY" );
//    if ( delay )
//    {
//        strcat( command_string, " export PSC_STRATEGY_DELAY=" );
//        strcat( command_string, delay );
//        strcat( command_string, ";" );
//    }
//
    sprintf( command_string, "%s %s", command_string, xstr( _AAGENT_EXEC ) );


    strcat( command_string, " --tag=" );
    strcat( command_string, agent->tag );

    strcat( command_string, " --parent=" );
    strcat( command_string, agent->parent );

    if( opts.has_phase ) {
        strcat( command_string, " --phase=" );
        strcat( command_string, opts.phase_string );
    }

    strcat( command_string, " --port=" );
    strcat( command_string, agent->port );

    strcat( command_string, " --appname=" );
    strcat( command_string, fe->get_appname() );

    strcat( command_string, " --id=" );
    strcat( command_string, agent->applIds );

    sprintf( buf, " --mpinumprocs=%d --ompnumthreads=%d", fe->get_mpinumprocs(), fe->get_ompnumthreads() );
    strcat( command_string, buf );


    if( opts.has_strategy ) {
        strcat( command_string, " --strategy=" );
        strcat( command_string, opts.strategy );
    }


    if( opts.has_delay ) {
        strcat( command_string, " --delay=" );
        strcat( command_string, opts.delay_string );
    }


    if( opts.has_registry ) {
        sprintf( buf, " --registry=%s", opts.reg_string );
        strcat( command_string, buf );
    }

    if( opts.has_timeout ) {
        sprintf( buf, " --timeout=%d", atoi( opts.timeout_string ) );
        strcat( command_string, buf );
    }

    sprintf( buf, " --debug=%d ", psc_dbg_level );
    strcat( command_string, buf );

    if( opts.has_inst ) {
        sprintf( buf, " --inst=%s", opts.inst_string );
        strcat( command_string, buf );
    }

    if( opts.has_selectivedebug ) {
        sprintf( buf, " --selective-debug=%s", opts.selectivedebug_string );
        strcat( command_string, buf );
    }


    strcat( command_string, "\"" );
    // START analysis agent..
    psc_dbgmsg( 1, "Starting AAgent (%s): %s\n", agent->tag, command_string );
    if( fork() == 0 ) {
        int code = system( command_string );
        // TODO handle the return code properly here
        exit( 0 );
    }
}


void AltixPBSStarter::runHlAgent( PBSAgentDetails* agent ) {
    char command_string[ 10000 ];
    char buf[ 1000 ];
    int  hostNumber = getHostNumber( agent->host );


    sprintf( command_string, "/usr/pbs/default/bin/pbsdsh -n %i -- /usr/pbs/default/bin/pbs_attach "
             "-j $PBS_JOBID /bin/bash -c \"export USER=$PBS_O_LOGNAME; export LOGNAME=$USER; "
             "export PBS_NODEFILE=/var/spool/PBS/aux/$PBS_JOBID; . /etc/profile; ",
             pbsnodelist_firstidx[ hostNumber ] );

    sprintf( command_string, "%s %s", command_string, xstr( _HLAGENT_EXEC ) );

    strcat( command_string, " --tag=" );
    strcat( command_string, agent->tag );


    strcat( command_string, " --parent=" );
    strcat( command_string, agent->parent );

    strcat( command_string, " --appname=" );
    strcat( command_string, fe->get_appname() );

    strcat( command_string, " --port=" );
    strcat( command_string, agent->port );

    strcat( command_string, " --children=" );
    strcat( command_string, agent->children );

    if( opts.has_registry ) {
        sprintf( buf, " --registry=%s", opts.reg_string );
        strcat( command_string, buf );
    }

    if( opts.has_dontcluster ) {
        strcat( command_string, " --dontcluster " );
    }

    if( opts.has_timeout ) {
        sprintf( buf, " --timeout=%d", atoi( opts.timeout_string ) );
        strcat( command_string, buf );
    }

    sprintf( buf, " --debug=%d ", psc_dbg_level );
    strcat( command_string, buf );

    if( opts.has_selectivedebug ) {
        sprintf( buf, " --selective-debug=%s", opts.selectivedebug_string );
        strcat( command_string, buf );
    }


    strcat( command_string, "\"" );

    // START high-level agent..
    psc_dbgmsg( 1, "Starting HLAgent (%s): %s\n", agent->tag, command_string );
    if( fork() == 0 ) {
        int code = system( command_string );
        // TODO handle the return code properly here
        exit( 0 );
    }
}

bool AltixPBSStarter::runAgents() {
    // child tag
    char child_tag[ 2000 ];

    sprintf( child_tag, "fe[%d]:0", fe->own_id() );

    char hostname[ 100 ];
    gethostname( hostname, 100 );
    fe->add_child_agent( child_tag, hostname, -1 );

    PBSLevelInfo* masterLevel = computeAgentHierarchy();
    printAgentHierarchy( masterLevel );

    PBSLevelInfo* level = masterLevel;
    while( level != NULL && level->prevLevel != NULL ) {
        PBSAgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runHlAgent( agent );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }

    if( level != NULL ) {
        PBSAgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runAnalysisAgent( agent );
            agent = agent->nextAgent;
        }
    }
//    char             command_string[ 8000 ];
//    char             buf[ 8000 ];
//    int              cores_per_processor = 4;
//    char             senvbuf[ 512 ];
//    int              max_runs = 4;
//    char             masteragent_placement[ 32768 ];
//    char             sbuf[ 32768 ];
//    char             appstart_cmd[ 32768 ];
//    char*            senvptr_old = NULL;
//    RegistryService* regsrv = fe->get_registry();
//
//    char tmpb[ 8192 ];
//
//
//    //
//    // find out hostname for master agent
//    //
//
//    std::list< std::string >::iterator nlit;
//    char                               masterhost[ 2000 ];
//
//    if ( opts.has_masterhost )
//    {
//        strcpy( masterhost, opts.masterhost_string );
//    }
//    else
//    {
//        gethostname( masterhost, 2000 );
//    }
//
//    int  node_hnumber = 0;
//    bool nnfound      = false;
////    for ( nlit = pbsnodelist.begin(); nlit != pbsnodelist.end(); nlit++ )
////    {
////      if ( !strcmp( masterhost,( *nlit ).c_str() ) )
////      {
////          nnfound = true;
////          break;
////      }
////      node_hnumber++;
////    }
//    if ( nnfound )
//    {
//        sprintf( command_string,
//                 "/usr/pbs/bin/pbsdsh -v -n %i -- /usr/pbs/bin/pbs_attach -j $PBS_JOBID /bin/bash -c \"export USER=$PBS_O_LOGNAME; "
//                 "export LOGNAME=$USER; export PBS_NODEFILE=/var/spool/PBS/aux/$PBS_JOBID; . /etc/profile; "
//                 "hlagent ",
//                 node_hnumber );
//    }
//    else
//    {
//      psc_errmsg( "masterhost %s not found!\n", masterhost );
//      exit( 1 );
//    }
//
//    psc_dbgmsg( 1, "querying registry for appname >%s<\n", fe->get_appname() );
//
//    std::list< EntryData >           result;
//    std::list< EntryData >::iterator entryit;
//    EntryData                        query;
//
//    query.comp = "MRIMONITOR";
//    query.app = fe->get_appname();
//    query.tag = "none";
//
//    if ( regsrv->query_entries(result, query) == -1 )
//    {
//        psc_errmsg( "Error querying registry for application\n" );
//        exit(1);
//    }
//    if( result.size() == 0 )
//    {
//        psc_errmsg( "Error querying registry for application\n" );
//        exit( 1 );
//    }
//
//    psc_dbgmsg( 1, "query to reg...ok, size: %d\n", result.size() );
//
//    std::map< std::string, std::list<int> >           apphosts;
//    std::map< std::string, std::list<int> >::iterator apphostit;
//    for( entryit = result.begin(); entryit != result.end(); entryit++ )
//    {
//        // appids.push_back( entryit->id );
//        apphosts[ entryit->node ].push_back( entryit->id );
//    }
//
//    std::list< int >::iterator idit;
//    char                       masteragent_apphosts[ 16384 ];
//
//    int idmap_h_cpu[ 16 ][ 512 ];
//    for ( int ctb = 0; ctb < 16; ctb++ )
//    {
//        for ( int ctc = 0; ctc < 512; ctc++ )
//        {
//            idmap_h_cpu[ ctb ][ ctc ] = -1;
//        }
//    }
//
//    psc_dbgmsg(1,"%d nodes, IDs per node:\n",apphosts.size());
//    bool foundany=false;
//    for (int ctt=0; ctt<8192; ctt++) {
//        idmap_f[ctt]=-1;
//        idmap_t[ctt]=-1;
//    }
//    idmap_len=0;
//    sprintf( masteragent_apphosts, " --apphosts=" );
//    for( apphostit=apphosts.begin(); apphostit!=apphosts.end(); apphostit++ ) {
//        psc_dbgmsg(1,"  for node %s there are %d IDs\n",(*apphostit).first.c_str(),(*apphostit).second.size());
//        if( apphostit!=apphosts.begin() ) {
//            strcat( masteragent_apphosts, ",");
//        }
//
//        int sfound_at=-1;
//        for (int cct=0; cct<pbsnodelist_size; cct++) {
//            if (pbsnodelist_string[cct]==(*apphostit).first.c_str()) {
//                sfound_at=cct;
//                break;
//            }
//        }
//
//        if (sfound_at<0) {
//            psc_errmsg("Could not find host >%s<\n",(*apphostit).first.c_str());
//            continue;
//        }
//
//        strcat( masteragent_apphosts, (*apphostit).first.c_str() );
//
//        for( idit=(*apphostit).second.begin(); idit!=(*apphostit).second.end(); idit++ ) {
//            sprintf(sbuf, ":%d", (*idit) );
//            strcat( masteragent_apphosts, sbuf);
//
//            EntryData entry;
//
//            if( regsrv->get_entry( entry, *idit, true )==-1 ) {
//                psc_errmsg("Error querying data for entry %d\n", *idit);
//                continue;
//            }
//
//            std::list<std::string>::iterator it;
//            int num=1;
//            for( it=entry.strings.begin(); it!=entry.strings.end(); it++ ) {
////                            psc_dbgmsg( 3, "String %d: %s\n", num++, it->c_str() );
//                int maxrank;
//                int myrank;
//                int mycpu;
//                if( sscanf( it->c_str(),
//                    "PERISCOPE processrank [%d:%d:%d]",
//                    &maxrank, &myrank, &mycpu ) > 0 ) {
//                    foundany=true;
//                    if (myrank>idmap_len) idmap_len=myrank;
//                    if (maxrank>idmap_len) idmap_len=maxrank;
//                    idmap_t[myrank]=*idit;
//
//                    if (idmap_h_cpu[sfound_at][mycpu]>=0) {
//                        psc_errmsg("slot %d:%d already contains %d, requested %d\n",
//                            sfound_at,mycpu,idmap_h_cpu[sfound_at][mycpu],*idit);
//                    } else {
//                        idmap_h_cpu[sfound_at][mycpu]=*idit;
//                    }
//
//                    int inum=idmap_h_num[sfound_at];
//                    idmap_h[sfound_at][inum]=*idit;
//                    idmap_h_num[sfound_at]=inum+1;
//                    psc_dbgmsg(3,"procrank %d for ID %d at [%d:%d], cpu %d\n",myrank,*idit,sfound_at,inum,mycpu);
//                }
//            }
//        }
//    }
//
//
//
//    //
//    // High-level agent hosts
//    //
//    char hostname[100];
//    gethostname(hostname, 100);
//    sprintf(buf, " --hlagenthosts=%s", hostname );
//    strcat( command_string, buf );
//
//
//    // placement for node agents under MPI-applications
//    sprintf(masteragent_placement," --placement=\"");
//    psc_dbgmsg(1,"pbsnodelist_size %d\n",pbsnodelist_size);
//    char ctemp[512];
//    for (int nct=0; nct<pbsnodelist_size; nct++) {
//        psc_dbgmsg(1,"Creating placement for host: %s\n",pbsnodelist_string[nct].c_str());
//        if (nct>0) strcat(masteragent_placement,".");
//        sprintf(ctemp,"#%s#",pbsnodelist_string[nct].c_str());
//        strcat(masteragent_placement,ctemp);
//
//        stg->create_placement_string(nct,masteragent_placement,threads,idmap_h_num[nct],&idmap_h_cpu[nct][0]);
//    }
//    strcat(masteragent_placement,"\"");
//    psc_dbgmsg(1,">%s<\n",masteragent_placement);
//
//
//    psc_dbgmsg(1, "read %d IDs\n",idmap_len);
//    for (int ct=0; ct<idmap_len; ct++) {
//        psc_dbgmsg(1, "ID %d at rank position %d\n",idmap_t[ct],ct);
//    }
//
////    char *args=generateStarterIndependentHLagentArguments();
////    sprintf(sbuf,"%s %s %s %s",command_string,masteragent_apphosts,masteragent_placement,args);
////    free(args);
//
//    // child tag
//    char child_tag[2000];
//
//    sprintf(child_tag, "fe[%d]:0", fe->own_id() );
//
//    strcat( sbuf, " --tag=" );
//    strcat( sbuf, child_tag );
//    strcat(sbuf, "\"");
//
//    fe->add_child_agent( child_tag, masterhost, -1 );
//
//    if( fork()==0 ) {
//        psc_dbgmsg(1, "Starting master agent now: >%s<\n",sbuf);
//
//        system( sbuf );
//        exit(0);
//    }

    return true;
}
