/**
   @file	BGP_Starter.cc
   @ingroup Frontend
   @brief   LL starter for BG/P systems
   @author	Yury Oleynik
   @verbatim
        Revision:       $Revision$
        Revision date:  $Date$
        Committed by:   $Author$

        This file is part of the Periscope performance measurement tool.
        See http://www.lrr.in.tum.de/periscope for details.

        Copyright (c) 2005-2011, Technische Universitaet Muenchen, Germany
        See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/types.h>  // Process id type return by fork(): pid_t
#include <math.h>
#include <unistd.h>

#include "frontend.h"
#include "BGP_Starter.h"
#include "regxx.h"
#include "psc_errmsg.h"
#include "psc_config.h"


#define SMALL_JOB_SIZE_LIMIT                    512
#define IO_RATIO_FZJ_MIN                        32
#define IO_RATIO_FZJ_MAX                        128
#define IO_RATIO_KAUST_MAX              128
#define IO_RATIO_KAUST_MIN              64

#pragma GCC diagnostic ignored "-Wwrite-strings"
using namespace std;

BGP_Starter::BGP_Starter() : Starter() {
    // TODO Auto-generated constructor stub
}

BGP_Starter::~BGP_Starter() {
    // TODO Auto-generated destructor stub
}

bool BGP_Starter::runApplication() {
    //creating the LL script

    FILE* fpout;
    pid_t pid = getpid();

    char* psc_home;
    if( ( psc_home = getenv( "PERISCOPE_ROOT" ) ) == 0 ) {
        psc_errmsg( "PERISCOPE_ROOT is not set can't continue..." );
        fe->quit();
    }

    char mode[ 100 ];
    int  job_factor = -11;
    if( opts.has_bg_mode ) {
        sprintf( mode, "-mode %s", opts.bg_mode );
        if( strcmp( opts.bg_mode, "VN" ) == 0 ) {
            job_factor = 4;
        }
        if( strcmp( opts.bg_mode, "DUAL" ) == 0 ) {
            job_factor = 2;
        }
        if( strcmp( opts.bg_mode, "SMP" ) == 0 ) {
            job_factor = 1;
        }
        if( job_factor == -1 ) {
            psc_errmsg( "BlueGene mode %s is not recognoized!\n", opts.bg_mode );
            fe->quit();
        }
    }

    std::string str_strategy = opts.strategy;
    if( job_factor > 1 && ( strcmp( str_strategy.c_str(), "BGP_Cache" ) == 0 ) ) {
        psc_errmsg( "BlueGene P Cache strategy is available only in SMP mode!\n", opts.bg_mode );
        fe->quit();
    }

    if( fe->get_mpinumprocs() % job_factor != 0 ) {
        psc_errmsg( "Number of requested mpi processes %d is not consistent with Bluegne mode %s!\n", fe->get_mpinumprocs(), opts.bg_mode );
        fe->quit();
    }

    int num_nodes = ( int )fe->get_mpinumprocs() / job_factor;

    //system("cp ./.periscope ./.periscope_reserv");
    int code = system( "head -n 7 .periscope >.periscope_clean" );
    code = system( "mv .periscope_clean .periscope" );
    // TODO handle these return codes

    if( ( fpout = fopen( "./.periscope", "a" ) ) == NULL ) {
        exit( 1 );
    }

    fprintf( fpout, "DEBUG=%d\n", psc_get_debug_level() );
    fprintf( fpout, "STRATEGY=%s\n", opts.strategy );

    if( strcmp( fe->sitename(), "KAUST" ) == 0 ) {
        fprintf( fpout, "PROCS_PER_AA=%d\n", IO_RATIO_KAUST_MAX * job_factor );
    }
    else if( strcmp( fe->sitename(), "FZJ" ) == 0 ) {
        fprintf( fpout, "PROCS_PER_AA=%d\n", IO_RATIO_FZJ_MAX * job_factor );
    }
    else {
        psc_errmsg( "Installation site is not supported\n", fe->get_mpinumprocs(), opts.bg_mode );
        fe->quit();
    }
    fprintf( fpout, "BG_MODE=%d\n", job_factor );
    fprintf( fpout, "AA_PARENT=bottomHL\n" );
    fprintf( fpout, "AA_TAG=AAgent\n" );
    fprintf( fpout, "PROCS=%d\n", fe->get_mpinumprocs() );
    fprintf( fpout, "APPNAME=%s\n", fe->get_appname() );
    fprintf( fpout, "OMPNUMTHREADS=%d\n", threads );
    fclose( fpout );

    if( psc_get_debug_level() == 160 ) {
        //system("cp ./.periscope ./.periscope_lastrun");
        //system("mv ./.periscope_reserv ./.periscope");
        return true;
    }


    if( opts.has_force_localhost ) {
        if( ( fpout = fopen( "psc_LLscript", "w" ) ) == NULL ) {
            exit( 1 );
        }

        fprintf( fpout, "# @ job_name = Periscope\n" );
        fprintf( fpout, "# @ comment = \"Performance analysis with Periscope\"\n" );
        fprintf( fpout, "# @ error = $(job_name).$(host).$(jobid).out\n" );
        fprintf( fpout, "# @ output = $(job_name).$(host).$(jobid).out\n" );
        fprintf( fpout, "# @ environment = COPY_ALL;\n" );
        fprintf( fpout, "# @ wall_clock_limit = 00:30:00\n" );
        fprintf( fpout, "# @ notification = never\n" );
        fprintf( fpout, "# @ notify_user = oleynik@in.tum.de\n" );
        fprintf( fpout, "# @ job_type = bluegene\n" );
        fprintf( fpout, "# @ bg_size = %d\n", num_nodes );
        fprintf( fpout, "# @ queue\n" );
        fprintf( fpout, "echo \"=================================================================\"\n" );

        if( threads > 1 ) {
            fprintf( fpout, "mpirun -start_tool %s/bin/psc_analysisagent %s -np %d -env OMP_NUM_THREADS=%d %s \n", psc_home, mode, processes, threads, opts.app_run_string );
        }
        else {
            fprintf( fpout, "mpirun -start_tool %s/bin/psc_analysisagent %s -np %d %s \n", psc_home, mode, processes, opts.app_run_string );
        }

        fclose( fpout );


        if( psc_get_debug_level() != 160 ) {
            //BGP_DEBUG dismiss
            int code = system( "llsubmit psc_LLscript" );
            // TODO handle the error code
        }
    }
    else {
        char command_string[ 10000 ];
        if( threads > 1 ) {
            sprintf( command_string, "mpirun -start_tool %s/bin/psc_analysisagent %s -np %d -env OMP_NUM_THREADS=%d %s \n", psc_home, mode, processes, threads, opts.app_run_string );
        }
        else {
            sprintf( command_string, "mpirun -start_tool %s/bin/psc_analysisagent %s -np %d %s \n", psc_home, mode, processes, opts.app_run_string );
        }
        pid_t pid;
        pid = fork();
        if( pid == 0 ) {
            // Executed by the child process
            psc_dbgmsg( 1, "Starting application in interactively: >%s<\n", command_string );
            int retVal = system( command_string );
            // Terminate the child process
            _exit( retVal );            // call kernel level cleanup and not the user level one!
        }
        else if( pid < 0 ) {
            psc_errmsg( "Error forking child process!\n" );
            return false;
        }

        //sprintf(command_string,"mpirun -start_tool %s/bin/psc_analysisagent %s -np %d %s &\n",psc_home,mode,processes,opts.app_run_string);
        //system(command_string);
    }



    return true;
}

bool BGP_Starter::rerunApplication() {
    return true;
}

void BGP_Starter::runAnalysisAgent( AgentDetails* agent ) {
}

void BGP_Starter::runHlAgent( AgentDetails* agent ) {
    char command_string[ 10000 ];
    char buf[ 1000 ];

    char* cwd = getenv( "PWD" );
    sprintf( command_string, "cd %s; %s", cwd, xstr( _HLAGENT_EXEC ) );

    strcat( command_string, " --tag=" );
    strcat( command_string, agent->tag );


    strcat( command_string, " --parent=" );
    strcat( command_string, agent->parent );

    strcat( command_string, " --appname=" );
    strcat( command_string, fe->get_appname() );

    strcat( command_string, " --port=" );
    strcat( command_string, agent->port );

    strcat( command_string, " --children=" );
    strcat( command_string, agent->children );

    if( opts.has_registry ) {
        sprintf( buf, " --registry=%s", opts.reg_string );
        strcat( command_string, buf );
    }

    if( opts.has_dontcluster ) {
        strcat( command_string, " --dontcluster " );
    }

    if( opts.has_timeout ) {
        sprintf( buf, " --timeout=%d", atoi( opts.timeout_string ) );
        strcat( command_string, buf );
    }

    sprintf( buf, " --debug=%d >HL", psc_dbg_level );
    strcat( command_string, buf );
    strcat( command_string, agent->port );
    strcat( command_string, " 2>&1 " );
    // START high-level agent..
    psc_dbgmsg( 1, "Starting HLAgent %s: %s\n", agent->tag, command_string );
    if( fork() == 0 ) {
        int code = system( command_string );
        // TODO handle the return code
        exit( 0 );
    }
}

void BGP_Starter::printAgentHierarchy( LevelInfo* levels ) {
    LevelInfo* level;

    level = levels;
    while( level != NULL ) {
        AgentDetails* agent;
        agent = level->agents;
        psc_dbgmsg( 1, "Agents on level: %d\n", level->numberOfAgents );
        while( agent != NULL ) {
            psc_dbgmsg( 1, "%s Port: %s, Host: %s, Parent: %s, ApplIds: %s, Children: %s\n",
                        agent->tag, agent->port, agent->host, agent->parent, agent->applIds, agent->children );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }
}

BGP_Starter::LevelInfo* BGP_Starter::computeAgentHierarchy() {
    //1. get number of app processes from frontend

    //2. compute the number of AA from the number of mpinumprocs:
    //			if mpinumprocs > 256			// mpinumprocs has to be multiple of 512
    //				numAA = mpinumprocs/IO_RATIO_2		// in SMP mode
    //				numAA = mpinumprocs/IO_RATIO_2*2		// in DUAL mode
    //				numAA = mpinumprocs/IO_RATIO_2*4		// in VN mode
    //			else							// mpinumprocs has to be multiple of IO_RATIO_1
    //				numAA = mpinumprocs/IO_RATIO_1          // in SMP mode
    //				numAA = mpinumprocs/IO_RATIO_1*2	// in DUAL mode
    //				numAA = mpinumprocs/IO_RATIO_1*4		// in VN mode
    //

    char hostname[ 100 ];
    strcpy( hostname, "localhost" );
    LevelInfo*    levels, * level, * masterLevel, * level2;
    AgentDetails* agent, * agent2;
    int           AAclusterSize       = 32;                                                             //Number of applProcesses under one AAgent
    int           L1ClusterSize       = fe->get_maxfan();                                               //Number of AAgents under 1st level of HlAgents
    int           maxfan              = fe->get_maxfan();
    int           total_agents_number = 0;

    char mode[ 100 ];
    int  job_factor = -11;
    if( opts.has_bg_mode ) {
        sprintf( mode, "-mode %s", opts.bg_mode );
        if( strcmp( opts.bg_mode, "VN" ) == 0 ) {
            job_factor = 4;
        }
        if( strcmp( opts.bg_mode, "DUAL" ) == 0 ) {
            job_factor = 2;
        }
        if( strcmp( opts.bg_mode, "SMP" ) == 0 ) {
            job_factor = 1;
        }
        if( job_factor == -1 ) {
            psc_errmsg( "BlueGene mode %s is not recognoized!\n", opts.bg_mode );
            fe->quit();
        }
    }

    if( fe->get_mpinumprocs() % job_factor != 0 ) {
        psc_errmsg( "Number of requested mpi processes %d is not consistent with Bluegne mode %s!\n", fe->get_mpinumprocs(), opts.bg_mode );
        fe->quit();
    }

    int num_apps = ( int )fe->get_mpinumprocs() / job_factor;   // number of compute nodes

    int num_AA;
    if( strcmp( fe->sitename(), "KAUST" ) == 0 ) {
        if( num_apps % IO_RATIO_KAUST_MIN != 0 ) {
            psc_dbgmsg( 6, "Selected number of compute nodes doesn't fulfill the requirements: it should be multiple of %d \n", IO_RATIO_KAUST_MIN );
            exit( 1 );
        }
//compute the number of AAs
        num_AA = num_apps / IO_RATIO_KAUST_MAX;
    }
    else if( strcmp( fe->sitename(), "FZJ" ) == 0 ) {
        if( num_apps % IO_RATIO_FZJ_MIN != 0 ) {
            psc_dbgmsg( 6, "Selected number of compute nodes doesn't fulfill the requirements: it should be multiple of %d \n", IO_RATIO_FZJ_MIN );
            exit( 1 );
        }
//compute the number of AAs
        num_AA = num_apps / IO_RATIO_FZJ_MAX;
    }
    else {
        psc_errmsg( "Installation site is not supported\n", fe->get_mpinumprocs(), opts.bg_mode );
        fe->quit();
    }

    /*

     */



    psc_dbgmsg( 6, "Approximate number of AAgents is %d \n", num_AA );

//create the first level of agents - AAgents

    levels = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level  = levels;

//create second level of HlAgents

    level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level->nextLevel->prevLevel = level;
    level2                      = level->nextLevel;

    int nextPort = fe->get_local_port() + 1;

//creating first AAgent

    agent                 = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
    level->agents         = agent;
    level->numberOfAgents = 1;
    psc_dbgmsg( 6, "AAgent 0 created.\n", agent2->port, agent2->host );

//creating first Level1 HL agent

    agent2                 = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
    level2->agents         = agent2;
    level2->numberOfAgents = 1;
    sprintf( agent2->port, "%d", nextPort );
    nextPort++;
    strcat( agent2->host, hostname );
    psc_dbgmsg( 6, "HlAgent 0 created: HL.port=%s,HL.host=%s!\n", agent2->port, agent2->host );

//creating AAgents and Level1 HLs

    int aa_count_per_hl_l1 = 1;
    for( int i = 1; i < num_AA; i++ ) {
        agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
        level->numberOfAgents++;
        agent = agent->nextAgent;
        psc_dbgmsg( 6, "AAgent %d created.\n", i );
        if( aa_count_per_hl_l1 >= L1ClusterSize ) {
            agent2->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            level2->numberOfAgents++;
            agent2 = agent2->nextAgent;
            sprintf( agent2->port, "%d", nextPort );
            nextPort++;
            strcat( agent2->host, hostname );
            aa_count_per_hl_l1 = 1;
            psc_dbgmsg( 6, "HlAgent %d created: HL.port=%s,HL.host=%s!\n", level2->numberOfAgents - 1, agent2->port, agent2->host );
        }
        else {
            aa_count_per_hl_l1++;
        }
    }


    level = level2;

//creating agents on the level starting from the second level of HLagents

    while( level->numberOfAgents > 1 ) {
        int numberOfAgentsInPrevLevel = level->numberOfAgents;

        level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
        level->nextLevel->prevLevel = level;
        level                       = level->nextLevel;
        agent                       = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
        sprintf( agent->port, "%d", nextPort );
        nextPort++;
        strcat( agent->host, hostname );


        level->agents         = agent;
        level->numberOfAgents = 1;
        for( int i = 1; i < ( numberOfAgentsInPrevLevel + maxfan - 1 ) / maxfan; i++ ) {
            agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent            = agent->nextAgent;
            sprintf( agent->port, "%d", nextPort );
            nextPort++;
            strcat( agent->host, hostname );

            level->numberOfAgents++;
        }
    }

    if( psc_get_debug_level() == 160 ) {
        psc_dbgmsg( 6, "\n\n*************************************\n*HERE IS THE TOPOLGY:               *\n*************************************\n\n" );
    }

// setting the tags for created agents

    masterLevel = level;
    int  bottomlevel = 0;
    char tagStr[ 100 ], parentStr[ 100 ], childrenTags[ 4000 ];

    sprintf( tagStr, "fe[%d]:0", fe->own_id() );

    if( level->prevLevel != NULL ) {
        if( level->prevLevel->prevLevel == NULL ) {
            sprintf( tagStr, "bottomHL" );
        }
    }

#ifndef _BGP_PORT_HEARTBEAT_V1
    gethostname( hostname, 100 );
    fe->add_child_agent( tagStr, hostname, -1 );
#endif

    strcat( level->agents->tag, tagStr );
    *childrenTags = 0;
    sprintf( parentStr, "%s[%d]", fe->get_local_tag().c_str(), fe->own_id() );
    strcat( level->agents->parent, parentStr );


    while( level->prevLevel != NULL ) {
        AgentDetails* child, * parent;
        int           childrenCnt = 0;
        parent = level->agents;
        child  = level->prevLevel->agents;

        if( level->prevLevel->prevLevel != NULL ) {
            if( level->prevLevel->prevLevel->prevLevel == NULL ) {
                bottomlevel = 1;
            }
        }
        else {
            bottomlevel = 2;
        }

        while( child != NULL && parent != NULL ) {
            sprintf( parentStr, "%s", parent->tag );
            strcat( child->parent, parentStr );


            switch( bottomlevel ) {
            case 0:
                sprintf( tagStr, "%s:%d", parent->tag, childrenCnt );
                //sprintf(tagStr, "highHL");
                break;
            case 1:
                sprintf( tagStr, "bottomHL" );
                break;
            case 2:
                sprintf( tagStr, "AAgent" );
                break;
            }


            strcat( child->tag, tagStr );
            strcat( childrenTags, tagStr );
            childrenCnt++;
            if( childrenCnt == maxfan ) {
                child       = child->nextAgent;
                childrenCnt = 0;
                strcat( parent->children, childrenTags );
                *childrenTags = 0;
                parent        = parent->nextAgent;
            }
            else {
                child = child->nextAgent;
                if( child != NULL ) {
                    strcat( childrenTags, "," );
                }
            }
        }

        if( parent != NULL ) {
            strcat( parent->children, childrenTags );
        }
        *childrenTags = 0;
        level         = level->prevLevel;
    }

// counting the total amount of agents

    level = masterLevel;
    while( level != NULL ) {
        total_agents_number += level->numberOfAgents;
        level                = level->prevLevel;
    }
    psc_dbgmsg( 6, "Total amount of agents is %d\n", total_agents_number );

    //BGP_DEBUG dismiss
    //fe->set_agents_number(total_agents_number-num_AA);
    fe->set_agents_number( total_agents_number );



    return masterLevel;
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}

bool BGP_Starter::runAgents() {
    char command_string[ 5000 ];
    char buf[ 1000 ];




    LevelInfo* masterLevel = computeAgentHierarchy();
    printAgentHierarchy( masterLevel );
    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////
    if( psc_get_debug_level() == 160 ) {
        return false;
    }
    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////


    //
    // start high-level agents
    //


    LevelInfo* level = masterLevel;
    while( level != NULL && level->prevLevel != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runHlAgent( agent );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }

    //if (level!=NULL){
    //	AgentDetails *agent=level->agents;
    //	while (agent!=NULL){
    //		runAnalysisAgent(agent);
    //		agent=agent->nextAgent;
    //	}
    //}
    return true;
}
