/**
   @file    FastInteractiveStarter.cc
   @ingroup Frontend
   @brief   Starter for interactive jobs (optimized for startup speed)
   @author  Isaias Compres
   @verbatim

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2013, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/types.h>
#include <unistd.h>
extern char** environ;

#include "frontend.h"
#include "FastInteractiveStarter.h"
#include "regxx.h"
#include "psc_errmsg.h"
#include "psc_config.h"
#include "selective_debug.h"
#include "config.h"

extern int  agent_pid;                          // this is an interactive launcher that sets the fast mode of operation
extern int  application_pid;                    // tracked in the frontend.cc
extern char user_specified_environment[ 5000 ]; // user provided environment variables for the starter

using namespace std;

FastInteractiveStarter::FastInteractiveStarter() : Starter() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to Starter()\n" );
}

FastInteractiveStarter::~FastInteractiveStarter() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to Starter()\n" );
}

bool FastInteractiveStarter::runApplication() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to runApplication()\n" );

    // TODO invert startup order: first AAgents followed by the MPI application -IC

    int child_status;

    processes = fe->get_mpinumprocs();
    char appstart_cmd[ 10000 ];
    char appstart_env_vars[ 5000 ];
    char agent_info_entry[ 1000 ];

    char mri_debug[ 1000 ] = "";

    // TODO remove hardcoded numbers here... -IC
    if( active_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( MRItrace ) ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneAll ) ) );
    }
    if( active_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneMRIMonitor ) ) ) {
        sprintf( mri_debug, "%d", psc_dbgLevel2mri_dbgLevel( PSC_SELECTIVE_DEBUG_LEVEL( AutotuneMRIMonitor ) ) );
    }
    // TODO verify what to do here with the ScoreP branch... -IC
    // TODO verify what to do with the KMP fields added in the other starters... -IC
    sprintf( appstart_env_vars, "export PSC_FASTMODE=true; export PSC_FRONTEND=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; export PSC_MRI_DEBUG=%s;",
             fe->get_local_hostname().c_str(), fe->get_local_port(), fe->get_ompnumthreads(), fe->get_appname(), mri_debug );

    char hostname[ 1024 ];
    hostname[ 1023 ] = '\0';
    gethostname( hostname, 1023 );

    // in this starter, the agent runs in the same host as the application (could use "localhost")
    // the port, however, needs to be obtained from the agent
    // TODO in the future, we may want to obtain the port of the agent from the frontend and pass it here, in order to run 2 or more agents -IC
    sprintf( agent_info_entry, " export PSC_AGENT%d=%s:%d;", 0, hostname, 51000 );
    strcat( appstart_env_vars, agent_info_entry );
    strcat( appstart_env_vars, " " );   // separate the options by at least 1 space
    strcat( appstart_env_vars, user_specified_environment );
    // TODO check what is done with the PSC_DEBUG_RUN variable here -IC
    // TODO add the different types of MPI commands here, if necessary (the standard recommends mpiexec but this is sometimes ignored) -IC
    sprintf( appstart_cmd, "%s mpiexec -n %d %s", appstart_env_vars, processes, opts.app_run_string );
    // TODO verify what was done with OMP threads here -IC
    // TODO double check what is going on here; this is the source of multiple frontend children when restarting -IC
    if( !applUninstrumented() ) {
        pid_t pid = fork();
        if( pid == 0 ) {
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ),
                        "Starting application in interactive partition: >%s<\n", appstart_cmd );
            // TODO check if this is the best way to start an MPI application; there are the exec variants available as well -IC
            // TODO change this to an execv call; need to rebuild the structures above
            // the system call passes its parameters to an unspecified shell; it is better to avoid the side-effects of running
            // on an unknown shell
            int retVal = system( appstart_cmd );
            _exit( retVal );            // TODO do we really want to call _exit? -IC
        }
        else if( pid < 0 ) {
            psc_errmsg( "Error forking child process!\n" );
            return false;
        }
        else {
            application_pid = pid;
        }
    }
    else {
        psc_dbgmsg( 1, "Starting non-instrumented application in interactive partition: >%s<\n", appstart_cmd );
        int retVal = system( appstart_cmd );
    }

    // TODO make sure the process book-keeping is done properly here or later, if necessary -IC
    // Here is what each process entry should include:
    // EntryData queryresult;
    // queryresult.app
    // queryresult.site
    // queryresult.mach
    // queryresult.id
    // queryresult.node
    // queryresult.port
    // queryresult.pid
    // queryresult.tag
    // queryresult.comp

    return true;
}

bool FastInteractiveStarter::rerunApplication() {
    RegistryService* regsrv = fe->get_registry();
    pid_t            pid;
    int              child_status;

    for( int rank = 0; rank < processes; rank++ ) {
        idmap_f[ rank ] = idmap_t[ rank ];
        idmap_t[ rank ] = -1;
    }

    int code =  runApplication();

    //send mapping info to child agents
    std::map< std::string, AgentInfo >*          child_agents = fe->get_child_agents();
    std::map< std::string, AgentInfo >::iterator it;
    for( it = ( *child_agents ).begin(); it != ( *child_agents ).end(); it++ ) {
        AgentInfo& ai = it->second;
        if( ai.status != AgentInfo::CONNECTED ) {
            if( fe->connect_to_child( &ai ) == -1 ) {
                psc_errmsg( "Error connecting to child at %s:%d\n", ai.hostname.c_str(), ai.port );
            }
            else {
                ai.handler->reinit( processes, idmap_f, idmap_t );
            }
        }
    }
    return code;
}

void FastInteractiveStarter::runAnalysisAgent( AgentDetails* agent ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to runAnalysisAgent()\n" );

    // TODO get proper maximums for these array declarations
    char  command[ 100 ];
    char  temp[ 1000 ];
    char* command_options[ 20 ]; // make sure this is the maximum number of options
    command_options[ 19 ] = 0;   // terminate as NULL for execv
    for( int i = 0; i < 19; i++ ) {
        command_options[ i ]      = ( char* )malloc( 100 * sizeof( char ) );
        command_options[ i ][ 0 ] = 0;
    }
    int command_options_count = 0;

    // agent command; full installed path
    sprintf( temp, "%s/bin/%s", PERISCOPE_INSTALL_DIRECTORY, xstr( _AAGENT_EXEC ) );
    strcpy( command, temp );
    // these are fast mode options
    strcpy( command_options[ command_options_count++ ], "psc_analysisagent" );
    strcpy( command_options[ command_options_count++ ], "--fastmode" );
    strcpy( command_options[ command_options_count++ ], "--leaf" ); // leaf nodes attach to application processes
    sprintf( temp, "--parent-port=%d", fe->get_local_port() );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--parent-host=%s", fe->get_local_hostname().c_str() );
    strcpy( command_options[ command_options_count++ ], temp );
    strcpy( command_options[ command_options_count++ ], "--dontregister" );
    sprintf( temp, "--tag=%s", agent->tag );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--parent=%s", agent->parent );
    strcpy( command_options[ command_options_count++ ], temp );

    if( opts.has_phase ) {
        sprintf( temp, "--phase=%s", opts.phase_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    sprintf( temp, "--port=%s", agent->port );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--appname=%s", fe->get_appname() );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--id=%s", agent->applIds );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--mpinumprocs=%d", fe->get_mpinumprocs() );
    strcpy( command_options[ command_options_count++ ], temp );
    sprintf( temp, "--ompnumthreads=%d", fe->get_ompnumthreads() );
    strcpy( command_options[ command_options_count++ ], temp );

    if( opts.has_strategy ) {
        sprintf( temp, "--strategy=%s", opts.strategy );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    if( opts.has_delay ) {
        sprintf( temp, "--delay=%s", opts.delay_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    if( opts.has_registry ) {
        sprintf( temp, "--registry=%s", opts.reg_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    if( opts.has_timeout ) {
        sprintf( temp, "--timeout=%s", opts.timeout_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    sprintf( temp, "--debug=%d", psc_dbg_level );
    strcpy( command_options[ command_options_count++ ], temp );

    if( opts.has_inst ) {
        sprintf( temp, "--inst=%s", opts.inst_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }

    if( opts.has_selectivedebug ) {
        sprintf( temp, "--selective-debug=%s", opts.selectivedebug_string );
        strcpy( command_options[ command_options_count++ ], temp );
    }
    command_options[ command_options_count ] = 0;

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Starting Analysis Agent: %s\n", agent->tag );
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Command: %s; option count: %d;\n", command, command_options_count );
    command_options_count = 0;
    while( command_options[ command_options_count ] != 0 ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "[%02d]: %s\n", command_options_count, command_options[ command_options_count ] );
        command_options_count++;
    }

    pid_t fork_code = fork();
    if( fork_code == 0 ) {    // child
        execv( command, command_options );
        psc_errmsg( "FastInteractiveStarter: execv (forked child) failed with the SHH command: %s; options: %d\n", command, command_options_count );
        throw 0;
    }
    else if( fork_code < 0 ) {     // could not fork child
        psc_errmsg( "FastInteractiveStarter: Error creating Analysis Agent with tag: %s.\n", agent->tag );
        throw 0;
    }
    else {       // at the parent; child pid returned
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "the child agent was started with PID: %d\n", ( int )fork_code );

        // the pid tracking method is only relevant in the interactive starters (FastInteractiveStarter and InteractiveStarter)
        // there is only one agent created when running in an interactive session
        agent_pid = fork_code;
    }
}

void FastInteractiveStarter::runHlAgent( AgentDetails* agent ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to runHlAgent()\n" );
    psc_errmsg( "FastInteractive never starts any High Level Agent! This event indicates a bug "
                "in FastInteractiveStarter::computeAgentHierarchy()\n" );
    throw 0;
}

void FastInteractiveStarter::printAgentHierarchy( LevelInfo* levels ) {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to printAgentHierarchy()\n" );
    // TODO this should be updated to an implementation specific version; it is currently copied from InteractiveStarter

    LevelInfo* level;
    int        levelNr = 1;
    level = levels;
    while( level != NULL ) {
        AgentDetails* agent;
        agent = level->agents;
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Agents on level %d: %d\n", levelNr++, level->numberOfAgents );
        while( agent != NULL ) {
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ),
                        "%s %s Port: %s; Host: %s; Parent: %s; ApplIds: %s; Children: %s;\n", ( level->prevLevel == NULL ) ? "[AA]" : "[HLA]",
                        agent->tag, agent->port, agent->host, agent->parent, agent->applIds,
                        ( strlen( agent->children ) ? agent->children : "none" ) );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }
}

FastInteractiveStarter::LevelInfo* FastInteractiveStarter::computeAgentHierarchy() {
    // TODO this seems like reusable code; evaluate if this is the case and refactor as necessary -IC
    // TODO this needs to be reworked for this starter (there is no need to compute any HL agent, for example). -IC
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to computeAgentHierarchy()\n" );
    RegistryService*                                  regsrv = fe->get_registry();
    std::list< EntryData >::iterator                  entryit;
    std::list< int >                                  appids;
    std::list< int >::iterator                        idit;
    std::map< std::string, std::list<int> >           apphosts;
    std::map< std::string, std::list<int> >::iterator apphostit;
    char                                              hostname[ 100 ];
    gethostname( hostname, 100 );
    LevelInfo*    levels, * level, * masterLevel;
    AgentDetails* agent;
    int           nextPort = fe->get_local_port() + 1;

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Identified %d matching entries\n", processList.size() );

    for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
        apphosts[ entryit->node ].push_back( entryit->id );
    }

    int clusterSize = 128;
    if( opts.has_maxcluster ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Maxcluster in interactive startup %s\n", opts.maxcluster_string );
        clusterSize = atoi( opts.maxcluster_string );
    }

    levels = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level  = levels;

    // Create an agent description
    agent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
    sprintf( agent->port, "%d", nextPort++ );
    strcat( agent->host, hostname );
    //determine analysis agents
    int   entryCount = 0;
    char* placement  = agent->applIds;
    *placement = 0;

    level->agents         = agent;
    level->numberOfAgents = 1;

    if( opts.has_force_localhost ) {
        for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
            entryCount++;
            //Split list of application entries into clusters of give size
            if( entryCount > clusterSize ) {
                //new analysis agent
                agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                agent            = agent->nextAgent;
                sprintf( agent->port, "%d", nextPort++ );
                strcat( agent->host, hostname );

                // set the pointer to the string that will store the appl ids and init the mem
                placement  = agent->applIds;
                *placement = 0;

                level->numberOfAgents++;
                entryCount = 1;

                // add the current app id to the new agent
                sprintf( strchr( placement, 0 ), "%d", entryit->id );
            }
            else {
                if( entryCount > 1 ) {
                    strcat( placement,  "," );
                }
                sprintf( strchr( placement, 0 ), "%d", entryit->id );
            }
        }
    }
    else {
        for( apphostit = apphosts.begin(); apphostit != apphosts.end(); apphostit++ ) {
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "Processing node %s\n", apphostit->first.c_str() );

            if( apphostit != apphosts.begin() ) {
                // create new agent
                agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                agent            = agent->nextAgent;
                sprintf( agent->port, "%d", nextPort++ );
                strcat( agent->host, hostname );

                // set the pointer to the string that will store the appl ids and init the mem
                placement  = agent->applIds;
                *placement = 0;

                level->numberOfAgents++;
                entryCount = 0;
            }

            // Loop over all process entries for the current node!
            for( idit = ( *apphostit ).second.begin(); idit != ( *apphostit ).second.end(); idit++ ) {
                entryCount++;
                //Split list of application entries into clusters of give size
                if( entryCount > clusterSize ) {
                    //new analysis agent
                    agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                    agent            = agent->nextAgent;
                    sprintf( agent->port, "%d", nextPort++ );
                    strcat( agent->host, hostname );

                    // set the pointer to the string that will store the appl ids and init the mem
                    placement  = agent->applIds;
                    *placement = 0;

                    level->numberOfAgents++;
                    entryCount = 1;

                    // add the current app id to the new agent
                    sprintf( strchr( placement, 0 ), "%d", ( *idit ) );
                }
                else {
                    if( entryCount > 1 ) {
                        strcat( placement,  "," );
                    }
                    sprintf( strchr( placement, 0 ), "%d", ( *idit ) );
                }
            }
        }
    }

    int maxfan = fe->get_maxfan();
    //Create levels of high level agents

    while( level->numberOfAgents > 1 ) {
        int numberOfAgentsInPrevLevel = level->numberOfAgents;

        // Create the next hierarchy level in the list of levels
        level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
        level->nextLevel->prevLevel = level;
        level                       = level->nextLevel;

        // HLA
        agent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
        sprintf( agent->port, "%d", nextPort++ );
        strcat( agent->host, hostname );
        level->agents         = agent;
        level->numberOfAgents = 1;

        // if maxfan is 1, then create only 1 HL for all agents
        if( maxfan == 1 ) {
            break;
        }

        for( int i = 1; i < ( numberOfAgentsInPrevLevel + maxfan - 1 ) / maxfan; i++ ) {
            agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent            = agent->nextAgent;
            sprintf( agent->port, "%d", nextPort );
            nextPort++;
            strcat( agent->host, hostname );
            level->numberOfAgents++;
        }
    }
    masterLevel = level;
    char tagStr[ 100 ], parentStr[ 100 ], childrenTags[ 4000 ];

    sprintf( tagStr, "fe[%d]:0", fe->own_id() );
    strcat( level->agents->tag, tagStr );
    *childrenTags = 0;
    sprintf( parentStr, "%s[%d]", fe->get_local_tag().c_str(), fe->own_id() );
    strcat( level->agents->parent, parentStr );
    while( level->prevLevel != NULL ) {
        AgentDetails* child, * parent;
        int           childrenCnt = 0;
        parent = level->agents;                 // List of agents on this level
        child  = level->prevLevel->agents;      // Children of the current level (agents on the next lower level)

        while( child != NULL && parent != NULL ) {
            sprintf( parentStr, "%s", parent->tag );
            strcat( child->parent, parentStr );

            sprintf( tagStr, "%s:%d", parent->tag, childrenCnt );
            strcat( child->tag, tagStr );

            strcat( childrenTags, tagStr );
            childrenCnt++;
            if( maxfan > 1 && childrenCnt == maxfan ) {
                childrenCnt = 0;
                strcat( parent->children, childrenTags );
                *childrenTags = 0;
                parent        = parent->nextAgent;
            }
            else {
                if( child != NULL ) {
                    strcat( childrenTags, "," );
                }
            }
            child = child->nextAgent;
        }

        if( parent != NULL ) {
            strcat( parent->children, childrenTags );
        }
        *childrenTags = 0;
        level         = level->prevLevel;
    }
    return masterLevel;
}

bool FastInteractiveStarter::runAgents() {
    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "call to runAgents()\n" );
    char command_string[ 5000 ];
    char buf[ 1000 ];

    // setting up the master agent first
    char child_tag[ 2000 ];
    sprintf( child_tag, "fe[%d]:0", fe->own_id() ); // only one connected to the fe
    char hostname[ 100 ];
    gethostname( hostname, 100 );                   // in this case, runs in the localhost
    fe->add_child_agent( child_tag, hostname, -1 );

    // the master agent's parent in the frontend
    char parent[ 2000 ];
    sprintf( parent, " --parent=%s:%d", fe->get_local_hostname().c_str(), fe->get_local_port() );

    char port[ 2000 ];
    // TODO should we get confirmation, which port it actually opened? -IC
    // this port is acquired from the frontend; there should be an Ack in the heartbeat, giving the actual opened one to be
    // passed to the application
    sprintf( port, "%d", fe->get_local_port() + 1 ); // tell it to open the next port

    LevelInfo* masterLevel = computeAgentHierarchy();
    printAgentHierarchy( masterLevel );

    // launching the high-level agents
    LevelInfo* level = masterLevel;
    while( level != NULL && level->prevLevel != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runHlAgent( agent );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }

    psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( HierarchySetup ), "launching individual agents...\n" );
    if( level != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runAnalysisAgent( agent );
            agent = agent->nextAgent;
        }
    }

    return true;
}
