/**
   @file	P6Starter.cc
   @ingroup Frontend
   @brief   POE starter for IBM Power575 nodes
   @author	Yury Oleynik
   @verbatim
        Revision:       $Revision$
        Revision date:  $Date$
        Committed by:   $Author$

        This file is part of the Periscope performance measurement tool.
        See http://www.lrr.in.tum.de/periscope for details.

        Copyright (c) 2005-2011, Technische Universitaet Muenchen, Germany
        See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>  // Process id type return by fork(): pid_t
#include <math.h>

#include "frontend.h"
#include "P6Starter.h"
#include "regxx.h"
#include "psc_errmsg.h"
#include "psc_config.h"

#pragma GCC diagnostic ignored "-Wwrite-strings"


using namespace std;

P6Starter::P6Starter() : Starter() {
    // TODO Auto-generated constructor stub
}

P6Starter::~P6Starter() {
    // TODO Auto-generated destructor stub
}

bool P6Starter::runApplication() {
    RegistryService* regsrv = fe->get_registry();
    pid_t            pid;
    int              child_status;

    if( psc_get_debug_level() == 160 ) {
        return true;
    }

    if( !opts.has_apprun ) {
        return true;
    }
    else {
        char appstart_cmd[ 5000 ];
#ifdef __p575
        pid = getpid();
        char output_dir[ 100 ];
        sprintf( output_dir, "periscope_run_%d", pid );
        int status = mkdir( output_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH );

        if( psc_get_debug_level() >= 6 ) {
            sprintf( appstart_cmd, "export PSC_REGISTRY=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; poe %s -procs %d 2>&1 > periscope_run_%d/AppDebugOutput", regsrv->get_reghost(), regsrv->get_regport(), fe->get_ompnumthreads(), fe->get_appname(), opts.app_run_string, processes, pid );
        }
        else {
            sprintf( appstart_cmd, "export PSC_REGISTRY=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; poe %s -procs %d ", regsrv->get_reghost(), regsrv->get_regport(), fe->get_ompnumthreads(), fe->get_appname(), opts.app_run_string, processes );
        }
#else
        sprintf( appstart_cmd, "export PSC_REGISTRY=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; mpirun -np %d %s",
                 regsrv->get_reghost(), regsrv->get_regport(), fe->get_ompnumthreads(), fe->get_appname(), processes, opts.app_run_string );
#endif
        if( threads > 0 ) {
            char buf[ 1000 ];
            sprintf( buf, "MPI_OMP_NUM_THREADS=%i", threads );
            putenv( buf );
            putenv( "MPI_OPENMP_INTEROP=yes" );
        }

        pid = fork();
        if( pid == 0 ) {
            // Executed by the child process
            psc_dbgmsg( 1, "Starting application in interactive partition: >%s<\n", appstart_cmd );
            int retVal = system( appstart_cmd );
            // Terminate the child process
            _exit( retVal );            // call kernel level cleanup and not the user level one!
        }
        else if( pid < 0 ) {
            psc_errmsg( "Error forking child process!\n" );
            return false;
        }

        EntryData query;

        bool startup_finished       = false;
        int  num_startup_components = processes;
        int  delay_interval_dt      = 3;

        if( num_startup_components == 0 ) {
            num_startup_components = 1;
            if( threads > 32 ) {
                delay_interval_dt = 10;
            }
        }
        else {
            if( processes > 64 ) {
                delay_interval_dt = 20;
            }
            else if( processes > 16 ) {
                delay_interval_dt = 10;
            }
        }

        psc_dbgmsg( 5, "waiting for startup of %d components...\n", num_startup_components );
        while( !startup_finished ) {
            query.app  = fe->get_appname();
            query.comp = "MRIMONITOR";
            query.tag  = "none";

            processList.clear();

            if( regsrv->query_entries( processList, query, false ) == -1 ) {
                psc_errmsg( "Error querying registry for application\n" );
                exit( 1 );
            }

            psc_dbgmsg( 1, "components started up: %d (of %d)\n", processList.size(), num_startup_components );
            startup_finished = ( processList.size() >= num_startup_components );

            // Child finished but components not started -> process was terminated or error occured
            if( ( waitpid( pid, &child_status, WNOHANG ) == pid ) && ( !startup_finished ) ) {
                return false;
            }

            if( !startup_finished ) {
                sleep( delay_interval_dt );
            }
        }         //waiting for all processes to register

        for( int pr = 0; pr < processes; pr++ ) {
            idmap_f[ pr ] = -1;
            idmap_t[ pr ] = -1;
        }

        std::list< EntryData >::iterator entryit;
        for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
            EntryData entry;
            entry                    = *entryit;
            idmap_t[ entry.pid - 1 ] = entry.id;
        }

        for( int i = 0; i < processes; i++ ) {
            if( idmap_t[ i ] == -1 ) {
                psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n", i );
                abort();
            }
        }
    }     //start application with mpirun
    return true;
}

bool P6Starter::rerunApplication() {
    RegistryService* regsrv = fe->get_registry();
    pid_t            pid;
    int              child_status;

    for( int rank = 0; rank < processes; rank++ ) {
        idmap_f[ rank ] = idmap_t[ rank ];
        idmap_t[ rank ] = -1;
    }

/**
 * Set time-out to the maximum value and call reinstrumetation procedure on restart
 */
    fe->set_global_timeout( 429496729 );
    instrumentRequiredRegions();

    char appstart_cmd[ 5000 ];
#ifdef __p575
    pid = getpid();
    sprintf( appstart_cmd, "export PSC_REGISTRY=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; poe %s -procs %d 2>&1 > periscope_run_%d/AppDebugOutput", regsrv->get_reghost(), regsrv->get_regport(), fe->get_ompnumthreads(), fe->get_appname(), opts.app_run_string, processes, pid );

#else
    sprintf( appstart_cmd, "export PSC_REGISTRY=%s:%d; export OMP_NUM_THREADS=%d; export PSC_APPNAME=%s; mpirun -np %d %s", regsrv->get_reghost(),
             regsrv->get_regport(), fe->get_ompnumthreads(), fe->get_appname(), processes, opts.app_run_string );
#endif

    if( threads > 0 ) {
        char buf[ 1000 ];
        sprintf( buf, "MPI_OMP_NUM_THREADS=%i", threads );
        putenv( buf );
        putenv( "MPI_OPENMP_INTEROP=yes" );
    }

    pid = fork();
    if( pid == 0 ) {
        // Executed by the child process
        psc_dbgmsg( 1, "Restarting application in interactive partition: >%s<\n", appstart_cmd );
        int retVal = system( appstart_cmd );
        // Terminate the child process
        _exit( retVal );        // call kernel level cleanup and not the user level one!
    }
    else if( pid < 0 ) {
        psc_errmsg( "Error forking child process!\n" );
        return false;
    }

    EntryData query;

    bool startup_finished  = false;
    int  delay_interval_dt = 2;


    if( processes > 16 ) {
        delay_interval_dt = 4;
    }
    else if( processes > 64 ) {
        delay_interval_dt = 8;
    }
    else if( processes > 1024 ) {
        delay_interval_dt = 15;
    }

    psc_dbgmsg( 5, "waiting for startup of %d components...\n", processes );
    while( !startup_finished ) {
        sleep( delay_interval_dt );
        query.app  = fe->get_appname();
        query.comp = "MRIMONITOR";
        query.tag  = "none";

        processList.clear();

        if( regsrv->query_entries( processList, query, false ) == -1 ) {
            psc_errmsg( "Error querying registry for application\n" );
            exit( 1 );
        }

        psc_dbgmsg( 1, "components started up: %d (of %d)\n", processList.size(), processes );
        startup_finished = ( processList.size() >= processes );

        // Child finished but components not started -> process was terminated or error occured
        if( ( waitpid( pid, &child_status, WNOHANG ) == pid ) && ( !startup_finished ) ) {
            return false;
        }
    }     //waiting for all processes to register

    std::list< EntryData >::iterator entryit;
    for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
        EntryData entry;
        entry                    = *entryit;
        idmap_t[ entry.pid - 1 ] = ( *entryit ).id;


//		std::list<std::string>::iterator stringit;
//		for (stringit=entry.strings.begin();stringit!=entry.strings.end(); stringit++){
//			int maxrank; int myrank; int mycpu;
//			if( sscanf( stringit->c_str(), "PERISCOPE processrank [%d:%d:%d]",
//				&maxrank, &myrank, &mycpu ) > 0 ) {
//				//psc_dbgmsg(2,"Entry %d: node %s, rank %d, mycpu %d, port %d\n", (*entryit).id, (*entryit).node.c_str(), myrank, mycpu, (*entryit).port);
//					  idmap_t[myrank]=(*entryit).id;
//			}
//		}
    }

    for( int i = 0; i < processes; i++ ) {
        if( idmap_t[ i ] == -1 ) {
            psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n", i );
            abort();
        }
    }

    //send mapping info to child agents
    std::map< std::string, AgentInfo >*          child_agents = fe->get_child_agents();
    std::map< std::string, AgentInfo >::iterator it;
    for( it = ( *child_agents ).begin(); it != ( *child_agents ).end(); it++ ) {
        AgentInfo& ai = it->second;
        if( ai.status != AgentInfo::CONNECTED ) {
            if( fe->connect_to_child( &ai ) == -1 ) {
                psc_errmsg( "Error connecting to child at %s:%d\n",
                            ai.hostname.c_str(), ai.port );
            }
            else {
                ai.handler->reinit( processes, idmap_f, idmap_t );
            }
        }
    }
    return true;
}

void P6Starter::runAnalysisAgent( AgentDetails* agent ) {
    char  command_string[ 10000 ];
    char  buf[ 1000 ];
    pid_t pid = getpid();

    /// \bug The agents should start in a working folder.. not in $HOME
    char* cwd = getenv( "PWD" );
    //sprintf( command_string,"ssh %s \"cd %s; ", agent->host, cwd);
    //sprintf( command_string,"poe ", agent->host, cwd);
//      char *delay = getenv("PSC_STRATEGY_DELAY");
//	if (delay){
//		strcat(command_string, " export PSC_STRATEGY_DELAY=");
//		strcat(command_string, delay);
//		strcat(command_string, ";");
//	}
    sprintf( command_string, "cd %s; ", cwd );
    strcat( command_string, xstr( _AAGENT_EXEC ) );

    strcat( command_string, " --tag=" );
    strcat( command_string, agent->tag );

    strcat( command_string, " --parent=" );
    strcat( command_string, agent->parent );

    if( opts.has_phase ) {
        strcat( command_string, " --phase=" );
        strcat( command_string, opts.phase_string );
    }

    strcat( command_string, " --port=" );
    strcat( command_string, agent->port );

    strcat( command_string, " --appname=" );
    strcat( command_string, fe->get_appname() );

    strcat( command_string, " --id=" );
    strcat( command_string, agent->applIds );

    sprintf( buf, " --mpinumprocs=%d --ompnumthreads=%d", fe->get_mpinumprocs(), fe->get_ompnumthreads() );
    strcat( command_string, buf );

    if( opts.has_strategy ) {
        strcat( command_string, " --strategy=" );
        strcat( command_string, opts.strategy );
    }

    if( opts.has_delay ) {
        strcat( command_string, " --delay=" );
        strcat( command_string, opts.delay_string );
    }

    if( opts.has_registry ) {
        sprintf( buf, " --registry=%s", opts.reg_string );
        strcat( command_string, buf );
    }
    if( opts.has_devControl ) {
        sprintf( buf, " --with-deviation-control" );
        strcat( command_string, buf );
    }

    if( opts.has_timeout ) {
        sprintf( buf, " --timeout=%d", atoi( opts.timeout_string ) );
        strcat( command_string, buf );
    }

    if( opts.has_inst ) {
        sprintf( buf, " --inst=%s", opts.inst_string );
        strcat( command_string, buf );
    }

//P575 DEBUG
    sprintf( buf, " --debug=%d > periscope_run_%d/AA", psc_dbg_level, pid );
    strcat( command_string, buf );
    strcat( command_string, agent->port );

//P575 DEBUG
    // Add the final "
    strcat( command_string, " 2>&1 " );

    // START analysis agent..
    psc_dbgmsg( 1, "Starting AAgent %s: %s\n", agent->tag, command_string );
    if( fork() == 0 ) {
        int code = system( command_string );
        // TODO handle the error code
        exit( 0 );
    }
}

void P6Starter::runHlAgent( AgentDetails* agent ) {
    char  command_string[ 10000 ];
    char  buf[ 1000 ];
    pid_t pid = getpid();

    char* cwd = getenv( "PWD" );
    sprintf( command_string, "cd %s; %s", cwd, xstr( _HLAGENT_EXEC ) );

    strcat( command_string, " --tag=" );
    strcat( command_string, agent->tag );


    strcat( command_string, " --parent=" );
    strcat( command_string, agent->parent );

    strcat( command_string, " --appname=" );
    strcat( command_string, fe->get_appname() );

    strcat( command_string, " --port=" );
    strcat( command_string, agent->port );

    strcat( command_string, " --children=" );
    strcat( command_string, agent->children );

    if( opts.has_registry ) {
        sprintf( buf, " --registry=%s", opts.reg_string );
        strcat( command_string, buf );
    }

    if( opts.has_dontcluster ) {
        strcat( command_string, " --dontcluster " );
    }

    if( opts.has_timeout ) {
        sprintf( buf, " --timeout=%d", atoi( opts.timeout_string ) );
        strcat( command_string, buf );
    }

    sprintf( buf, " --debug=%d > periscope_run_%d/HL", psc_dbg_level, pid );
    strcat( command_string, buf );
    strcat( command_string, agent->port );
    strcat( command_string, " 2>&1 " );
    // START high-level agent..
    psc_dbgmsg( 1, "Starting HLAgent %s: %s\n", agent->tag, command_string );
    if( fork() == 0 ) {
        int code = system( command_string );
        // TODO handle the error code
        exit( 0 );
    }
}

void P6Starter::printAgentHierarchy( LevelInfo* levels ) {
    LevelInfo* level;

    level = levels;
    while( level != NULL ) {
        AgentDetails* agent;
        agent = level->agents;
        psc_dbgmsg( 1, "Agents on level: %d\n", level->numberOfAgents );
        while( agent != NULL ) {
            psc_dbgmsg( 1, "%s Port: %s, Host: %s, Parent: %s, ApplIds: %s, Children: %s\n",
                        agent->tag, agent->port, agent->host, agent->parent, agent->applIds, agent->children );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }
}

P6Starter::LevelInfo* P6Starter::computeAgentHierarchy() {
    RegistryService*                        regsrv = fe->get_registry();
    std::list< EntryData >::iterator        entryit;
    std::list< int >                        appids;
    std::list< int >::iterator              idit;
    std::map< std::string, std::list<int> > apphosts;


    std::list<std::list<int> >           applIDsList;
    std::list<std::string>               hostList;
    std::list<std::list<int> >::iterator applIDsList_it;
    std::list<std::string>::iterator     hostList_it;


    std::map< std::string, std::list<int> >::iterator apphostit;
    char                                              hostname[ 100 ];
    gethostname( hostname, 100 );
    LevelInfo*    levels, * level, * masterLevel, * level2;
    AgentDetails* agent, * agent2;
    int           nextPort = fe->get_local_port() + 1;

//	query.app = fe->get_appname();
//	query.tag = "none";
//
////	if( regsrv->query_entries(result, query)==-1 ){
//		psc_errmsg( "Error querying registry for application\n");
//		exit(1);
//	}
//	if( result.size()==0 ){
//		psc_errmsg( "Could not find any entries with application name %s.\n", fe->get_appname() );
//		std::cout << query.app << query.comp << std::endl;
//		exit(1);
//	} else {
    psc_dbgmsg( 3, "Identified %d matching entries\n", processList.size() );

    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////


    /*for (int i=0;i<3;i++){
            for(int j=0;j<32;j++){
                    EntryData applentry;
                    char nodeName[10];
                    sprintf(nodeName,"node%d",i);
                    applentry.id=i*32+j;
                    applentry.app="appl";
                    applentry.node=nodeName;
                    applentry.port=i*32+j;
                    processList.push_back(applentry);
                    psc_dbgmsg(6,"pushing back %s id=%d, node=%s\n",applentry.app.c_str(),applentry.id,applentry.node.c_str());
            }
       }*/
    int nodeCount = 0;
    int nodeSize  = 32;
    if( psc_get_debug_level() == 160 ) {
        processList.clear();
        for( int i = 0; i < processes; i++ ) {
            EntryData applentry;
            char      nodeName[ 10 ];
            nodeCount = ( int )i / nodeSize;
            sprintf( nodeName, "node%d", nodeCount );
            applentry.id   = i;
            applentry.app  = "appl";
            applentry.node = nodeName;
            applentry.port = i;
            processList.push_back( applentry );
            psc_dbgmsg( 6, "pushing back %s id=%d, node=%s\n", applentry.app.c_str(), applentry.id, applentry.node.c_str() );
        }
    }
    nodeCount = ceil( ( double )processList.size() / ( double )nodeSize );
    psc_dbgmsg( 6, "number of procs=%d nodeCount=%d\n", ( int )processList.size(), nodeCount );
    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////

    //create a map from node names to ApplIDs list
    for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
        apphosts[ entryit->node ].push_back( entryit->id );
    }

    //split the map into two associated lists of hostnames and ApplIDs lists
    //and sort them together according to the size of ApplIDs list
    //Warning! here only one iteration of sort is performed, so it is assumed that all
    //nodes but one have equal size and only one node is less(incomplete)
    int MaxApplCountperNode = ( *apphosts.begin() ).second.size();
    for( apphostit = apphosts.begin(); apphostit != apphosts.end(); apphostit++ ) {
        if( ( *apphostit ).second.size() < MaxApplCountperNode ) {
            applIDsList.push_back( ( *apphostit ).second );
            hostList.push_back( ( *apphostit ).first );
        }
        else {
            applIDsList.push_front( ( *apphostit ).second );
            hostList.push_front( ( *apphostit ).first );
            MaxApplCountperNode = ( *apphostit ).second.size();
        }
    }

    //for (applIDsList_it=applIDsList.begin(),hostList_it=hostList.begin();applIDsList_it!=applIDsList.end();hostList_it++,applIDsList_it++){
    //	for (idit=(*applIDsList_it).begin();idit!=(*applIDsList_it).end();idit++){
    //		psc_dbgmsg(6,"applID=%d at %s\n",(*idit),(*hostList_it).c_str());
    //	}
    //}



    //create the first level of agents - AAgents
    levels = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level  = levels;


    //create second level of HlAgents
    level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level->nextLevel->prevLevel = level;
    level2                      = level->nextLevel;


    //get clusterSize as it was provided. Currently is not used, cluster size is determined by AAclusterSize
    int clusterSize = 4;
    if( opts.has_maxcluster ) {
        psc_dbgmsg( 5, "Maxcluster in interactive startup %s\n", opts.maxcluster_string );
        clusterSize = atoi( opts.maxcluster_string );
    }

    int AAclusterSize = 4;   //Number of applProcesses under one AAgent
    int L1ClusterSize = 4;   //Number of AAgents under 1st level of HlAgents

    char* placement;
    int   entryCount = 0;
    int   AAcount    = 0;

    bool firstTime = true;

    //determine analysis agents
    //Split list of application entries into clusters of given size
    //since we are on p575, we iterate over the hosts of application and
    //split them into clusters with one AA for each

    //loop over application's nodes
    //for (apphostit=apphosts.begin(); apphostit!=apphosts.end(); apphostit++){
    for( applIDsList_it = applIDsList.begin(), hostList_it = hostList.begin(); applIDsList_it != applIDsList.end(); hostList_it++, applIDsList_it++ ) {
        sprintf( hostname, "%s", ( *hostList_it ).c_str() );
        psc_dbgmsg( 5, "current host is %s\n", hostname );


        //if it is not the first Agent on the AA level then create one (the first one has been already created)
        if( firstTime ) {
            agent                  = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent2                 = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            level->agents          = agent;
            level->numberOfAgents  = 0;
            level2->agents         = agent2;
            level2->numberOfAgents = 0;
            firstTime              = false;
        }
        else {
            agent->nextAgent  = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent             = agent->nextAgent;
            agent2->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent2            = agent2->nextAgent;
        }

        //create first AAgent for the new node
        level->numberOfAgents++;                                                                                                                                                        ///?????????????????????????
        sprintf( agent->port, "%d", nextPort );
        nextPort++;
        strcat( agent->host, hostname );
        placement  = agent->applIds;
        entryCount = 0;

        psc_dbgmsg( 6, "AAgent cluster created: AA.port=%s,AA.host=%s AA.applIds=%s!\n", agent->port, agent->host, agent->applIds );

        //create first HLagent for the new node
        level2->numberOfAgents++;                                                                                                                                                       /// ?????????????????????????
        sprintf( agent2->port, "%d", nextPort );
        nextPort++;
        strcat( agent2->host, hostname );
        AAcount = 1;
        psc_dbgmsg( 6, "HlAgent cluster created: HL.port=%s,HL.host=%s!\n", agent2->port, agent2->host );


        //loop over application IDs inside one node
        //for( idit=(*apphostit).second.begin(); idit!=(*apphostit).second.end(); idit++ ){
        for( idit = ( *applIDsList_it ).begin(); idit != ( *applIDsList_it ).end(); idit++ ) {
            // if AA cluster is full then create new AA and assign this ID to the AA
            if( entryCount >= AAclusterSize ) {
                psc_dbgmsg( 6, "AAgent cluster full: AA.port=%s,AA.host=%s AA.applIds=%s, creating new one...\n", agent->port, agent->host, agent->applIds );
                char buf[ 100 ];

                //strcat(placement, ".");
                //new analysis agent

                agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                level->numberOfAgents++;
                agent = agent->nextAgent;
                sprintf( agent->port, "%d", nextPort );
                nextPort++;
                strcat( agent->host, hostname );
                placement = agent->applIds;
                sprintf( buf, "%d:0", ( *idit ) );
                strcat( placement, buf );
                entryCount = 1;
                psc_dbgmsg( 6, "AAgent cluster created: AA.port=%s,AA.host=%s AA.applIds=%s!\n", agent->port, agent->host, agent->applIds );

                //if HL cluster is full create new HL, later created above AA will go under the new HL
                if( AAcount >= L1ClusterSize ) {
                    psc_dbgmsg( 6, "HlAgent cluster full: HL.port=%s,HL.host=%s, creating new one...\n", agent2->port, agent2->host );
                    agent2->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                    level2->numberOfAgents++;
                    agent2 = agent2->nextAgent;
                    sprintf( agent2->port, "%d", nextPort );
                    nextPort++;
                    strcat( agent2->host, hostname );
                    AAcount = 1;
                    psc_dbgmsg( 6, "HlAgent cluster created: HL.port=%s,HL.host=%s!\n", agent2->port, agent2->host );
                }
                else {
                    AAcount++;
                }
            }
            else {
                char buf[ 100 ];

                if( entryCount == 0 ) {
                    sprintf( buf, "%d:0", ( *idit ) );
                }
                else {
                    sprintf( buf, ",%d:0", ( *idit ) );
                }
                strcat( placement, buf );
                entryCount++;
            }
            psc_dbgmsg( 5, "current applId is %d\n", ( *idit ) );
        }
    }
    psc_dbgmsg( 6, "one cluster ready AA.port=%s,AA.host=%s AA.applIds=%s\n", agent->port, agent->host, agent->applIds );
    strcat( placement, "" );

    int maxfan = fe->get_maxfan();


    //creating hierarchy starting from level 2 which go directly to additional nodes
    level = level2;
    int AgentsPerExtraNode = 1;
    nodeSize = 32;

    //if we have less then 24 application processes then we can run everything on one node
    if( processList.size() <= nodeSize ) {
        sprintf( hostname, "%s", ( *hostList.begin() ).c_str() );
    }
    else {
        sprintf( hostname, "extraNode%d", nodeCount );
    }
    //Create levels of high level agents
    while( level->numberOfAgents > 1 ) {
        int numberOfAgentsInPrevLevel = level->numberOfAgents;

        level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
        level->nextLevel->prevLevel = level;
        level                       = level->nextLevel;
        agent                       = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
        sprintf( agent->port, "%d", nextPort );
        nextPort++;
        strcat( agent->host, hostname );

        //increase number of agents on the current extraNode, if the node is full switch to the next one
        if( AgentsPerExtraNode >= nodeSize ) {
            nodeCount++;
            sprintf( hostname, "extraNode%d", nodeCount );
            AgentsPerExtraNode = 1;
        }
        else {
            AgentsPerExtraNode++;
        }

        level->agents         = agent;
        level->numberOfAgents = 1;
        for( int i = 1; i < ( numberOfAgentsInPrevLevel + maxfan - 1 ) / maxfan; i++ ) {
            agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent            = agent->nextAgent;
            sprintf( agent->port, "%d", nextPort );
            nextPort++;
            strcat( agent->host, hostname );

            //increase number of agents on the current extraNode, if the node is full switch to the next one

            if( AgentsPerExtraNode >= nodeSize ) {
                nodeCount++;
                sprintf( hostname, "extraNode%d", nodeCount );
                AgentsPerExtraNode = 1;
            }
            else {
                AgentsPerExtraNode++;
            }

            level->numberOfAgents++;
        }
    }

    if( psc_get_debug_level() == 160 ) {
        psc_dbgmsg( 6, "\n\n*************************************\n*TOTAL AMOUNT OF NODES REQUIRED = %d *\n*************************************\n\n", nodeCount + 1 );
    }

    masterLevel = level;

    char tagStr[ 100 ], parentStr[ 100 ], childrenTags[ 4000 ];

    sprintf( tagStr, "fe[%d]:0", fe->own_id() );
    strcat( level->agents->tag, tagStr );
    *childrenTags = 0;
    sprintf( parentStr, "%s[%d]", fe->get_local_tag().c_str(), fe->own_id() );
    strcat( level->agents->parent, parentStr );
    while( level->prevLevel != NULL ) {
        AgentDetails* child, * parent;
        int           childrenCnt = 0;
        parent = level->agents;
        child  = level->prevLevel->agents;
        while( child != NULL && parent != NULL ) {
            sprintf( parentStr, "%s", parent->tag );
            strcat( child->parent, parentStr );
            sprintf( tagStr, "%s:%d", parent->tag, childrenCnt );
            strcat( child->tag, tagStr );
            strcat( childrenTags, tagStr );
            childrenCnt++;
            if( childrenCnt == maxfan ) {
                child       = child->nextAgent;
                childrenCnt = 0;
                strcat( parent->children, childrenTags );
                *childrenTags = 0;
                parent        = parent->nextAgent;
            }
            else {
                child = child->nextAgent;
                if( child != NULL ) {
                    strcat( childrenTags, "," );
                }
            }
        }
        if( parent != NULL ) {
            strcat( parent->children, childrenTags );
        }
        *childrenTags = 0;
        level         = level->prevLevel;
        if( level->prevLevel != NULL ) {
            if( level->prevLevel->prevLevel == NULL ) {
                maxfan = L1ClusterSize;
            }
        }
    }
    return masterLevel;
}

bool P6Starter::runAgents() {
    char command_string[ 5000 ];
    char buf[ 1000 ];


    // child tag
    char child_tag[ 2000 ];

    sprintf( child_tag, "fe[%d]:0", fe->own_id() );

    char hostname[ 100 ];
    gethostname( hostname, 100 );
    fe->add_child_agent( child_tag, hostname, -1 );


    //
    // parent host and port
    //
    char parent[ 2000 ];
    sprintf( parent, " --parent=%s:%d", fe->get_local_hostname().c_str(),
             fe->get_local_port() );

    //
    // hlagent port
    //
    char port[ 2000 ];
    sprintf( port, "%d", fe->get_local_port() + 1 );

    //start analysis agent
    int clusterSize = 4;
    int entryCount;
    int analysisAgentNode = 1;

    char placement[ 2000 ];

    LevelInfo* masterLevel = computeAgentHierarchy();
    printAgentHierarchy( masterLevel );
    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////
    if( psc_get_debug_level() == 160 ) {
        return false;
    }
    ///////////////////////////////////////////////////////////////
    ///////////////////////////////p575 debug LLstarter development
    ///////////////////////////////////////////////////////////////


    //
    // start high-level agents
    //


    LevelInfo* level = masterLevel;
    while( level != NULL && level->prevLevel != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runHlAgent( agent );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }

    if( level != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runAnalysisAgent( agent );
            agent = agent->nextAgent;
        }
    }
    return true;
}
