/**
   @file    SLURMStarter.cc
   @ingroup Frontend
   @brief   Starter for SLURM interactive jobs / Generic MPI starter
   @author  Michael Gerndt
   @verbatim
    Revision:       $Revision$
    Revision date:  Sep 17, 2015
    Committed by:   Robert Mijakovic

    This file is part of the Periscope performance measurement tool.
    See http://www.lrr.in.tum.de/periscope for details.

    Copyright (c) 2005-2015, Technische Universitaet Muenchen, Germany
    See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/types.h> // Process id type return by fork(): pid_t

#include "frontend.h"
#include "SLURMStarter.h"
#include "regxx.h"
#include "psc_errmsg.h"
#include "psc_config.h"
#include "selective_debug.h"
#include "config.h"

extern int         application_pid;                    // tracked in the frontend.cc
extern char        user_specified_environment[ 5000 ]; // user provided environment variables for the starter
static const char* enopt_env_vars  = "EnoptTP,";
static const char* openmp_env_vars = "OpenMPTP;";
static const char* job_startup_cmd = "srun -n";

using namespace std;


SLURMStarter::SLURMStarter( void ) : Starter() {
}

SLURMStarter::~SLURMStarter( void ) {
}

bool SLURMStarter::runApplication( void ) {
    RegistryService* regsrv = fe->get_registry();
    int              child_status;

    processes = fe->get_mpinumprocs();
    if( !opts.has_apprun ) {
        return true;
    }
    else {
        std::stringstream appstart_env_vars;
        std::stringstream appstart_cmd;

        appstart_env_vars <<  "export SCOREP_ONLINEACCESS_ENABLE=true; export SCOREP_ENABLE_TRACING=false; export SCOREP_ONLINEACCESS_REG_PORT="
                          <<  regsrv->get_regport()
                          <<  "; export SCOREP_ONLINEACCESS_REG_HOST="  << regsrv->get_reghost()
                          <<  "; export SCOREP_ONLINEACCESS_APPL_NAME="  << fe->get_appname()
                          <<  "; export SCOREP_TUNING_PLUGINS="  <<  enopt_env_vars  <<  openmp_env_vars,
        printf( "%s\n", appstart_env_vars.str().c_str() );

        // set the environment as requested by the plugins
        appstart_env_vars <<  " "  <<  user_specified_environment;
        appstart_env_vars <<  "export OMP_NUM_THREADS="  <<  fe->get_ompnumthreads()  <<  "; ";
        appstart_cmd <<  appstart_env_vars.rdbuf()  <<  " "  <<  job_startup_cmd
                     <<  " "  <<  processes  <<  " "  <<  opts.app_run_string;

        if( !applUninstrumented() ) {
            /* the value of pid will be:
             *  negative - the creation of a child process was unsuccessful
             *  zero - returned in the addr space of the newly created child process
             *  positive - process ID of the child process (addr space of the parent)
             */
            pid_t pid = fork();
            if( pid == 0 ) {
                const std::string appstart_cmd_str = appstart_cmd.str();

                // Executed by the child process
                psc_dbgmsg( 1,
                            "Starting application in interactive partition: >%s<\n",
                            appstart_cmd_str.c_str() );
                int retVal = system( appstart_cmd_str.c_str() );
                // Terminate the child process
                _exit( retVal ); // call kernel level cleanup and not the user level one!
            }
            else if( pid < 0 ) {
                psc_errmsg( "Error forking child process!\n" );
                return false;
            }
            else {
                application_pid = pid;
            }

            EntryData query;

            bool startup_finished       = false;
            int  num_startup_components = processes;
            int  delay_interval_dt      = 3;

            if( num_startup_components == 0 ) {
                num_startup_components = 1;
                if( threads > 32 ) {
                    delay_interval_dt = 10;
                }
            }
            else {
                if( processes > 64 ) {
                    delay_interval_dt = 20;
                }
                else if( processes > 16 ) {
                    delay_interval_dt = 10;
                }
            }

            psc_dbgmsg( 5, "waiting for startup of %d components...\n",
                        num_startup_components );
            while( !startup_finished ) {
                query.app = fe->get_appname();
                query.tag = "none";

                processList.clear();

                if( regsrv->query_entries( processList, query, false ) == -1 ) {
                    psc_errmsg( "Error querying registry for application\n" );
                    abort();
                }

                psc_dbgmsg( 1, "components started up: %d (of %d)\n",
                            processList.size(), num_startup_components );
                startup_finished = ( processList.size() >= num_startup_components );

                // Child finished but components not started -> process was terminated or error occurred
                if( ( waitpid( pid, &child_status, WNOHANG ) == pid )
                    && ( !startup_finished ) ) {
                    return false;
                }

                if( !startup_finished ) {
                    sleep( delay_interval_dt );
                }
            } //waiting for all processes to register

            for( int pr = 0; pr < processes; pr++ ) {
                idmap_f[ pr ] = -1;
                idmap_t[ pr ] = -1;
            }

            std::list<EntryData>::iterator entryit;
            for( entryit = processList.begin(); entryit != processList.end();
                 entryit++ ) {
                EntryData entry = *entryit;

                if( entry.comp != "SCOREP" ) {
                    psc_errmsg( "Monitoring library %s is not supported!\n",
                                entry.comp.c_str() );
                    abort();
                }
                idmap_t[ entry.pid - 1 ] = entry.id;
            }

            for( int i = 0; i < processes; i++ ) {
                if( idmap_t[ i ] == -1 ) {
                    psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n",
                                i );
                    abort();
                }
            }
        }
        else {
            const std::string appstart_cmd_str = appstart_cmd.str();
            psc_dbgmsg( 1,
                        "Starting non-instrumented application in interactive partition: >%s<\n",
                        appstart_cmd_str.c_str() );
            int retVal = system( appstart_cmd_str.c_str() );
        }
    }
    return true;
}

bool SLURMStarter::rerunApplication( void ) {
    RegistryService* regsrv = fe->get_registry();
    pid_t            pid;
    int              child_status;

    for( int rank = 0; rank < processes; rank++ ) {
        idmap_f[ rank ] = idmap_t[ rank ];
        idmap_t[ rank ] = -1;
    }

    /**
     * Set time-out to the maximum value and call re-instrumentation procedure on restart
     */
    fe->set_global_timeout( 429496729 );
    instrumentRequiredRegions();
    std::stringstream appstart_cmd;
std: stringstream appstart_env_vars;

    appstart_env_vars <<  "export SCOREP_ONLINEACCESS_ENABLE=true; export SCOREP_ENABLE_TRACING=false; export SCOREP_ONLINEACCESS_REG_PORT="
                      <<  regsrv->get_regport()
                      <<  "; export SCOREP_ONLINEACCESS_REG_HOST="  <<  regsrv->get_reghost()
                      <<  "; export SCOREP_ONLINEACCESS_APPL_NAME="  <<  fe->get_appname()
                      <<  "; export SCOREP_TUNING_PLUGINS="  <<  enopt_env_vars  <<  openmp_env_vars
                      <<  "; export OMP_NUM_THREADS="  <<  fe->get_ompnumthreads();
    appstart_env_vars <<  " "  <<  user_specified_environment;

    appstart_cmd <<  appstart_env_vars.rdbuf()  <<  " "  << job_startup_cmd
                 <<  " "  <<  processes  <<  " "  <<  opts.app_run_string;

    pid = fork();
    if( pid == 0 ) {
        const std::string appstart_cmd_str = appstart_cmd.str();
        psc_dbgmsg( 1, "Restarting application in interactive partition: >%s<\n",
                    appstart_cmd_str.c_str() );
        int retVal = system( appstart_cmd_str.c_str() );
        _exit( retVal ); // call kernel level cleanup and not the user level one!
    }
    else if( pid < 0 ) {
        psc_errmsg( "Error forking child process!\n" );
        return false;
    }
    EntryData query;

    bool startup_finished  = false;
    int  delay_interval_dt = 2;

    if( processes > 16 ) {
        delay_interval_dt = 4;
    }
    else if( processes > 64 ) {
        delay_interval_dt = 8;
    }
    else if( processes > 1024 ) {
        delay_interval_dt = 15;
    }

    psc_dbgmsg( 5, "waiting for startup of %d components...\n", processes );
    while( !startup_finished ) {
        sleep( delay_interval_dt );
        query.app = fe->get_appname();
        query.tag = "none";

        processList.clear();

        if( regsrv->query_entries( processList, query, false ) == -1 ) {
            psc_errmsg( "Error querying registry for application\n" );
            abort();
        }

        psc_dbgmsg( 1, "components started up: %d (of %d)\n",
                    processList.size(), processes );
        startup_finished = ( processList.size() >= processes );

        // Child finished but components not started -> process was terminated or error occurred
        if( ( waitpid( pid, &child_status, WNOHANG ) == pid )
            && ( !startup_finished ) ) {
            return false;
        }
    } //waiting for all processes to register

    std::list<EntryData>::iterator entryit;
    for( entryit = processList.begin(); entryit != processList.end(); entryit++ ) {
        EntryData entry = *entryit;
        idmap_t[ entry.pid - 1 ] = ( *entryit ).id;

        if( entry.comp != "SCOREP" ) {
            psc_errmsg( "Monitoring library %s is not supported!\n",
                        entry.comp.c_str() );
            abort();
        }
    }

    for( int i = 0; i < processes; i++ ) {
        if( idmap_t[ i ] == -1 ) {
            psc_errmsg( "Error in collecting ids of processes. Registry entry for rank %d not found.\n",
                        i );
            abort();
        }
    }

    //send mapping info to child agents
    std::map< std::string, AgentInfo >*          child_agents = fe->get_child_agents();
    std::map< std::string, AgentInfo >::iterator it;
    for( it = ( *child_agents ).begin(); it != ( *child_agents ).end(); it++ ) {
        AgentInfo& ai = it->second;
        if( ai.status != AgentInfo::CONNECTED ) {
            if( fe->connect_to_child( &ai ) == -1 ) {
                psc_errmsg( "Error connecting to child at %s:%d\n",
                            ai.hostname.c_str(), ai.port );
            }
            else {
                ai.handler->reinit( processes, idmap_f, idmap_t );
            }
        }
    }

    return true;
}

void SLURMStarter::runAnalysisAgent( AgentDetails* agent ) {
    std::stringstream agent_args;
    std::stringstream pre_command;
    const char*       cwd = getenv( "PWD" );

    if( opts.has_force_localhost ) {
        pre_command <<  "cd "  <<  cwd  <<  ";";
    }
    else {
        // Use ssh to start the agent on remote hosts
        pre_command <<  "ssh "  <<  agent->host  <<  " \"cd "  <<  cwd  <<  "\";";
    }

    agent_args <<  " --tag="  <<  agent->tag;
    agent_args <<  " --parent="  <<  agent->parent;

    if( !opts.has_phase )
        psc_abort( "The name of the phase region was not provided!\n" );

    agent_args <<  " --phase="  <<  opts.phase_string;
    agent_args <<  " --port="  <<  agent->port;
    agent_args <<  " --appname="  <<  fe->get_appname();
    agent_args <<  " --id="  <<  agent->applIds;
    agent_args <<  " --mpinumprocs="  <<  fe->get_mpinumprocs()  <<  " --ompnumthreads="  <<  fe->get_ompnumthreads();
    if( opts.has_strategy )
        agent_args <<  " --strategy="  <<  opts.strategy;
    if( opts.has_delay )
        agent_args <<  " --delay="  <<  opts.delay_string;
    if( opts.has_registry )
        agent_args <<  " --registry="  <<  opts.reg_string;
    if( opts.has_timeout )
        agent_args <<  " --timeout="  <<  opts.timeout_string;
    agent_args <<  " --debug="  <<  psc_dbg_level;
    if( opts.has_inst )
        agent_args <<  " --inst="  <<  opts.inst_string;
    if( opts.has_selectivedebug )
        agent_args <<  " --selective-debug="  <<  opts.selectivedebug_string;
    if( !opts.has_force_localhost )
        agent_args <<  " \"";

    const char*       debug_tool = " gdb --nx --batch -ex run -ex bt -ex continue -ex quit --args ";
    std::stringstream command;
    command <<  pre_command.rdbuf()  <<  debug_tool  <<  ' '  <<  xstr(_AAGENT_EXEC)  <<  ' '  <<  agent_args.rdbuf();
    const std::string command_str = command.str();

    // START analysis agent..
    psc_dbgmsg( 1, "Starting AAgent (%s): %s\n", agent->tag, command_str.c_str() );
    if( fork() == 0 ) {
        int retVal = system( command_str.c_str() );
        _exit( retVal );
    }
}

void SLURMStarter::runHlAgent( AgentDetails* agent ) {
    std::stringstream command_string;

    if( opts.has_force_localhost )
        command_string <<  xstr( _HLAGENT_EXEC );
    else
        // Use ssh to start the agent on remote hosts
        command_string <<  "ssh "  <<  agent->host  <<  " "  <<  xstr( _HLAGENT_EXEC );

    command_string <<  " --tag="  <<  agent->tag;
    command_string <<  " --parent="  <<  agent->parent;
    command_string <<  " --appname="  <<  fe->get_appname();
    command_string <<  " --port="  <<  agent->port;
    command_string <<  " --child="  <<  agent->children;
    if( opts.has_registry )
        command_string <<  " --registry="  <<  opts.reg_string;
    if( opts.has_dontcluster )
        command_string <<  " --dontcluster";
    if( opts.has_timeout )
        command_string <<  " --timeout="  <<  atoi(opts.timeout_string);
    command_string <<  " --debug="  <<  psc_dbg_level;
    if( opts.has_selectivedebug )
        command_string <<  " --selective-debug="  <<  opts.selectivedebug_string;

    // START high-level agent..
    const std::string command = command_string.str();
    psc_dbgmsg( 1, "Starting HLAgent (%s): %s\n", agent->tag, command.c_str() );

    if( fork() == 0 ) {
        int retVal = system( command.c_str() );
        exit( retVal );
    }
}

void SLURMStarter::printAgentHierarchy( LevelInfo* levels ) {
    LevelInfo* level   = levels;
    int        levelNr = 1;
    while( level != NULL ) {
        AgentDetails* agent = level->agents;
        psc_dbgmsg( 1, "Agents on level %d: %d\n", levelNr++,
                    level->numberOfAgents );
        while( agent != NULL ) {
            psc_dbgmsg( 1,
                        "%s %s Port: %s; Host: %s; Parent: %s; ApplIds: %s; Children: %s;\n",
                        ( level->prevLevel == NULL ) ? "[AA]" : "[HLA]", agent->tag,
                        agent->port, agent->host, agent->parent, agent->applIds,
                        ( strlen( agent->children ) ? agent->children : "none" ) );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }
}

SLURMStarter::LevelInfo* SLURMStarter::computeAgentHierarchy( void ) {
    RegistryService*                                    regsrv = fe->get_registry();
    std::list<EntryData>::iterator                      entryit;
    std::list<int>                                      appids;
    std::list<int>::iterator                            idit;
    std::map< std::string, std::list< int > >           apphosts;
    std::map< std::string, std::list< int > >::iterator apphostit;
    char                                                hostname[ 100 ];
    gethostname( hostname, 100 );
    LevelInfo*    levels, * level, * masterLevel;
    AgentDetails* agent;
    int           nextPort = fe->get_local_port() + 1;

    psc_dbgmsg( 3, "Identified %d matching entries\n", processList.size() );

    for( entryit = processList.begin(); entryit != processList.end();
         entryit++ ) {
        apphosts[ entryit->node ].push_back( entryit->id );
    }

    int clusterSize = 128;
    if( opts.has_maxcluster ) {
        psc_dbgmsg( 3, "Maxcluster in interactive startup %s\n",
                    opts.maxcluster_string );
        clusterSize = atoi( opts.maxcluster_string );
    }

    levels = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
    level  = levels;

    // Create an agent description
    agent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
    sprintf( agent->port, "%d", nextPort++ );
    strcat( agent->host, hostname );
    //determine analysis agents
    int   entryCount = 0;
    char* placement  = agent->applIds;
    *placement = 0;

    level->agents         = agent;
    level->numberOfAgents = 1;

    if( opts.has_force_localhost ) {
        for( entryit = processList.begin(); entryit != processList.end();
             entryit++ ) {
            entryCount++;
            //Split list of application entries into clusters of give size
            if( entryCount > clusterSize ) {
                //new analysis agent
                agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                agent            = agent->nextAgent;
                sprintf( agent->port, "%d", nextPort++ );
                strcat( agent->host, hostname );

                // set the pointer to the string that will store the appl ids and init the mem
                placement  = agent->applIds;
                *placement = 0;

                level->numberOfAgents++;
                entryCount = 1;

                // add the current app id to the new agent
                sprintf( strchr( placement, 0 ), "%d", entryit->id );
            }
            else {
                if( entryCount > 1 ) {
                    strcat( placement, "," );
                }
                sprintf( strchr( placement, 0 ), "%d", entryit->id );
            }
        }
    }
    else {
        for( apphostit = apphosts.begin(); apphostit != apphosts.end();
             apphostit++ ) {
            psc_dbgmsg( 3, "Processing node %s\n", apphostit->first.c_str() );

            if( apphostit != apphosts.begin() ) {
                // create new agent
                agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                agent            = agent->nextAgent;
                sprintf( agent->port, "%d", nextPort++ );
                strcat( agent->host, hostname );

                // set the pointer to the string that will store the appl ids and init the mem
                placement  = agent->applIds;
                *placement = 0;

                level->numberOfAgents++;
                entryCount = 0;
            }

            // Loop over all process entries for the current node!
            for( idit = ( *apphostit ).second.begin();
                 idit != ( *apphostit ).second.end(); idit++ ) {
                entryCount++;
                //Split list of application entries into clusters of give size
                if( entryCount > clusterSize ) {
                    //new analysis agent
                    agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
                    agent            = agent->nextAgent;
                    sprintf( agent->port, "%d", nextPort++ );
                    strcat( agent->host, hostname );

                    // set the pointer to the string that will store the appl ids and init the mem
                    placement  = agent->applIds;
                    *placement = 0;

                    level->numberOfAgents++;
                    entryCount = 1;

                    // add the current app id to the new agent
                    sprintf( strchr( placement, 0 ), "%d", ( *idit ) );
                }
                else {
                    if( entryCount > 1 ) {
                        strcat( placement, "," );
                    }
                    sprintf( strchr( placement, 0 ), "%d", ( *idit ) );
                }
            }
        }
    }

    int maxfan = fe->get_maxfan();
    //Create levels of high level agents

    while( level->numberOfAgents > 1 ) {
        int numberOfAgentsInPrevLevel = level->numberOfAgents;

        // Create the next hierarchy level in the list of levels
        level->nextLevel            = ( LevelInfo* )calloc( 1, sizeof( LevelInfo ) );
        level->nextLevel->prevLevel = level;
        level                       = level->nextLevel;

        // HLA
        agent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
        sprintf( agent->port, "%d", nextPort++ );
        strcat( agent->host, hostname );
        level->agents         = agent;
        level->numberOfAgents = 1;

        // if maxfan is 1, then create only 1 HL for all agents
        if( maxfan == 1 ) {
            break;
        }

        for( int i = 1; i < ( numberOfAgentsInPrevLevel + maxfan - 1 ) / maxfan;
             i++ ) {
            agent->nextAgent = ( AgentDetails* )calloc( 1, sizeof( AgentDetails ) );
            agent            = agent->nextAgent;
            sprintf( agent->port, "%d", nextPort );
            nextPort++;
            strcat( agent->host, hostname );
            level->numberOfAgents++;
        }
    }
    masterLevel = level;
    char tagStr[ 100 ], parentStr[ 100 ], childrenTags[ 4000 ];

    sprintf( tagStr, "fe[%d]:0", fe->own_id() );
    strcat( level->agents->tag, tagStr );
    *childrenTags = 0;
    sprintf( parentStr, "%s[%d]", fe->get_local_tag().c_str(), fe->own_id() );
    strcat( level->agents->parent, parentStr );
    while( level->prevLevel != NULL ) {
        AgentDetails* child, * parent;
        int           childrenCnt = 0;
        parent = level->agents;            // List of agents on this level
        child  = level->prevLevel->agents; // Children of the current level (agents on the next lower level)

        while( child != NULL && parent != NULL ) {
            sprintf( parentStr, "%s", parent->tag );
            strcat( child->parent, parentStr );

            sprintf( tagStr, "%s:%d", parent->tag, childrenCnt );
            strcat( child->tag, tagStr );

            strcat( childrenTags, tagStr );
            childrenCnt++;
            if( maxfan > 1 && childrenCnt == maxfan ) {
                childrenCnt = 0;
                strcat( parent->children, childrenTags );
                *childrenTags = 0;
                parent        = parent->nextAgent;
            }
            else {
                if( child != NULL ) {
                    strcat( childrenTags, "," );
                }
            }
            child = child->nextAgent;
        }

        if( parent != NULL ) {
            strcat( parent->children, childrenTags );
        }
        *childrenTags = 0;
        level         = level->prevLevel;
    }
    return masterLevel;
}

bool SLURMStarter::runAgents( void ) {
    // child tag
    char child_tag[ 2000 ];

    sprintf( child_tag, "fe[%d]:0", fe->own_id() );

    char hostname[ 100 ];
    gethostname( hostname, 100 );
    fe->add_child_agent( child_tag, hostname, -1 );

    //
    // parent host and port
    //
    char parent[ 2000 ];
    sprintf( parent, " --parent=%s:%d",
             fe->get_local_hostname().c_str(), fe->get_local_port() );

    //
    // hlagent port
    //
    char port[ 2000 ];
    sprintf( port, "%d", fe->get_local_port() + 1 );

    //start analysis agent
    int clusterSize = 4;
    int entryCount;
    int analysisAgentNode = 1;

    char placement[ 2000 ];

    LevelInfo* masterLevel = computeAgentHierarchy();
    printAgentHierarchy( masterLevel );
    //
    // start high-level agents
    //

    LevelInfo* level = masterLevel;
    while( level != NULL && level->prevLevel != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runHlAgent( agent );
            agent = agent->nextAgent;
        }
        level = level->prevLevel;
    }

    if( level != NULL ) {
        AgentDetails* agent = level->agents;
        while( agent != NULL ) {
            runAnalysisAgent( agent );
            agent = agent->nextAgent;
        }
    }

    return true;
}
