/**
   @file    Starter.cc
   @ingroup Frontend
   @brief   Base class for front-end starters
   @author  Michael Gerndt
   @verbatim
   Revision:       $Revision$
   Revision date:  $Date$
   Committed by:   $Author$

   This file is part of the Periscope performance measurement tool.
   See http://www.lrr.in.tum.de/periscope for details.

   Copyright (c) 2005-2011, Technische Universitaet Muenchen, Germany
   See the COPYING file in the base directory of the package for details.
   @endverbatim
 */

#include "Starter.h"
#include "frontend.h"

using namespace std;

Starter::Starter() {
    executable = opts.app_string;
    processes  = fe->get_mpinumprocs();
    threads    = fe->get_ompnumthreads();
}

Starter::~Starter() {
}

std::string get_filename_from_region( std::string region ) {
    size_t      start, stop;
    std::string filename;

    start = region.find( "f: " );
    stop  = region.find( ",", start + 3 );
    if( region.find_last_of( "/", stop ) != string::npos ) {
        start = region.find_last_of( "/", stop );
        start = start - 2;
    }
    filename = region.substr( start + 3, stop - start - 3 );
    return filename;
}

std::string get_rfl_from_region( std::string region ) {
    size_t      start, stop;
    std::string rfl;

    start = region.find( "r: " );
    stop  = region.find( "\n", start + 3 );
    rfl   = region.substr( start + 3, stop - start - 3 );

    return rfl;
}


//TODO: MRI monitor code, should be adapted to Score-P if possible
void Starter::instrumentRequiredRegions() {
    bool        requiresMake = false;
    int         retVal       = 0;
    std::string make, src_folder, instr_folder;

    if( opts.has_inst != 1 ) {
        return;
    }

    fe->rm_duplicates_requiredRegionsList();
    fe->print_badRegions();
    fe->print_requiredRegionsList();
    fe->print_requiredRegionsListLastRun();

    if( opts.has_make == 1 ) {
        make = opts.make_string;
    }
    else {
        make = "make";
    }

    if( opts.has_sourcefolder == 1 ) {
        src_folder = opts.sourcefolder_string;
        src_folder.append( "/" );
    }
    else {
        src_folder = "./";
    }

    if( opts.has_instfolder == 1 ) {
        instr_folder = opts.instfolder_string;
        instr_folder.append( "/" );
    }
    else {
        instr_folder = "./inst/";
    }


    if( fe->get_badRegions().size() > 0 && !fe->get_badRegionsRemoved() &&
        ( strcmp( opts.inst_string, "overhead" ) == 0 || strcmp( opts.inst_string, "all_overhead" ) == 0 ) ) {
        size_t                           start, stop, startln, stopln;
        std::string                      filename, region, path, tmp, make, line;
        char                             cmnd[ 1000 ];
        std::list < std::string >        bad_regions = fe->get_badRegions();
        std::list<std::string>::iterator itr;

        tmp.erase();
        if( opts.has_config ) {
            tmp.append( opts.config_string );
        }
        else {
            tmp.append( src_folder );
            tmp.append( "psc_inst_config" );
        }
        sprintf( cmnd, "reinstrumentedall.sh %s", tmp.c_str() );
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
        retVal = system( cmnd );


        for( itr = bad_regions.begin(); itr != bad_regions.end(); itr++ ) {
            stop  = itr->find( ",", 0 );
            start = 2;
            if( itr->find_last_of( "/", stop ) != string::npos ) {
                start = itr->find_last_of( "/", stop );
            }
            filename = itr->substr( start + 1, stop - start - 1 );

            tmp = src_folder;
            tmp.append( filename );
            sprintf( cmnd, "touch %s", tmp.c_str() );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
            retVal = system( cmnd );

            stop     = itr->find_last_of( ".", stop );
            filename = itr->substr( start + 1, stop - start - 1 );
            filename.append( ".f90" );

            start  = itr->find( "r: ", stop );
            stop   = itr->find( "\n", start + 3 );
            region = itr->substr( start + 3, stop - start - 3 );

            //if (region != "0") {
            tmp = instr_folder;
            tmp.append( filename );
            sprintf( cmnd, "comment.sh %s %s ", region.c_str(), tmp.c_str() );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
            retVal = system( cmnd );
            //}
        }
        requiresMake = true;
        fe->set_badRegionsRemoved( true );
    }
    else if( strcmp( opts.inst_string, "analysis" ) == 0 ) {   //Analysis instrumentation strategy
        size_t                           start, stop, startln, stopln;
        std::string                      filename, region, path, tmp, make, line;
        char                             cmnd[ 1000 ];
        std::list < std::string >        bad_regions = fe->get_badRegions();
        std::list<std::string>::iterator itr;
        bool                             valid;


        if( !fe->get_allRegionsCommented() ) {
            tmp.erase();
            if( opts.has_config ) {
                tmp.append( opts.config_string );
            }
            else {
                tmp.append( src_folder );
                tmp.append( "psc_inst_config" );
            }

            sprintf( cmnd, "reinstrumentedall.sh %s", tmp.c_str() );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
            retVal = system( cmnd );

            sprintf( cmnd, "commentall.sh %s", instr_folder.c_str() );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
            retVal = system( cmnd );

            sprintf( cmnd, "touchall.sh %s", src_folder.c_str() );
            psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
            retVal = system( cmnd );

            requiresMake = true;
            fe->set_allRegionsCommented( true );
        }

        if( fe->checkForReinstrumentation() ) {       //There are required regions
            requiresMake = true;

            //First, remove the old one
            std::list < std::string > regList = fe->get_requiredRegionsListLastRun();
            for( std::list<std::string>::iterator it = regList.begin(); it != regList.end(); it++ ) {
                std::string region = *it;
                std::string filename, tmp, rfl;

                filename = get_filename_from_region( region );
                rfl      = get_rfl_from_region( region );

                tmp = src_folder;
                tmp.append( filename );
                sprintf( cmnd, "touch %s", tmp.c_str() );
                psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
                retVal = system( cmnd );

                stop     = filename.find_last_of( "." );
                filename = filename.substr( 0, stop );
                filename.append( ".f90" );

                tmp = instr_folder;
                tmp.append( filename );
                sprintf( cmnd, "comment.sh %s %s", rfl.c_str(), tmp.c_str() );
                psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
                retVal = system( cmnd );
            }

            //Second, instrument the required regions
            std::list<std::string> requiredRegList = fe->get_requiredRegionsList();
            for( std::list<std::string>::iterator it = requiredRegList.begin(); it != requiredRegList.end(); it++ ) {
                std::string region = *it;
                std::string filename, tmp, rfl;

                if( fe->is_in_badRegions( region ) ) {
                    continue;                     //This region would have too much overhead.
                }

                filename = get_filename_from_region( region );
                rfl      = get_rfl_from_region( region );

                tmp = src_folder;
                tmp.append( filename );
                sprintf( cmnd, "touch %s", tmp.c_str() );
                psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );
                retVal = system( cmnd );

                stop     = filename.find_last_of( "." );
                filename = filename.substr( 0, stop );
                filename.append( ".f90" );

                tmp = instr_folder;
                tmp.append( filename );
                sprintf( cmnd, "uncomment.sh %s %s", rfl.c_str(), tmp.c_str() );
                psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending command: %s to the system...\n", cmnd );

                retVal = system( cmnd );
            }
            fe->copy_requiredRegionsToLastRun();
            fe->clear_requiredRegionsList();
        } //New instrumentation required
        fe->clear_requiredRegionsList();
    }     //inst==analysis

    if( requiresMake ) {
        psc_dbgmsg( PSC_SELECTIVE_DEBUG_LEVEL( Autoinstrument ), "sending make command\n" );
        retVal = system( make.c_str() );
    }
}
